#!/usr/bin/env python3

import logging
import os
import pandas as pd
import sys
import shutil
import subprocess
import time

import drep
import drep.WorkDirectory
import drep.d_filter

def d_bonus_wrapper(wd,**kwargs):
    logging.info("Loading work directory")
    wd = drep.WorkDirectory.WorkDirectory(wd)
    logging.debug(str(wd))

    if kwargs.get('check_dependencies'):
        logging.info('Checking dependencies')
        check_dependencies()

    if kwargs.get('run_tax'):
        logging.info('Running tax')
        run_taxonomy(wd,**kwargs)

def run_taxonomy(wd, **kwargs):
    # Validate arguments- make sure you have everything you need
    Bdb, prod_dir, cent_dir = validate_arguments(wd, **kwargs)

    # Run prodigal
    drep.d_filter.run_prodigal(Bdb['location'].tolist(), prod_dir, **kwargs)

    # Run centrifuge
    run_centrifuge(Bdb, prod_dir, cent_dir, wd=wd, **kwargs)

    # # Call a centrifuge parser that returns Tdb
    # Tdb = parse_centrifuge(Bdb, cent_dir, **kwargs)
    #
    # # Add taxonomic info to Bdb
    # Bdb = add_taxonomy(Bdb,Tdb)

    # Parse taxonomy
    Tdb, Bdb = parse_taxonomy(Bdb, cent_dir, **kwargs)

    # Save Tdb and Bdb
    wd.store_db(Tdb,'Tdb',overwrite=True)
    wd.store_db(Bdb,'Bdb',overwrite=True)

def parse_taxonomy(Bdb, cent_dir, **kwargs):
    '''
    take the centrifuge directory and Bdb, return Tdb and Bdb with an added 'taxonomy' column
    '''

    method = kwargs.get('tax_method')

    if method == 'max':
        Tdb = parse_centrifuge(Bdb, cent_dir, **kwargs)

    elif method == 'percent':
        Tdb = parse_centrifuge_percent(Bdb, cent_dir, **kwargs)

    else:
        logging.error("dont recognize method {0}, quitting".format(method))
        sys.exit()

    Bdb = add_taxonomy(Bdb,Tdb)
    return Tdb, Bdb

def validate_arguments(wd, **kwargs):
    '''
    make sure you have everything you need
    '''
    if wd.hasDb('Bdb'):
        if kwargs.get('genomes',None) != None:
            logging.error("Both Bdb and a genome list are found- either don't include "\
                    + "a genome list or start a new work directory!")
            sys.exit()
        Bdb = wd.get_db('Bdb')

    else:
        if kwargs.get('genomes',None) == None:
            logging.error("I don't have anything to determine the taxonomy of! Give me a genome list")
            sys.exit()
        Bdb = drep.d_cluster.utils.load_genomes(kwargs['genomes'])

    prod_dir = wd.get_dir('prodigal')
    cent_dir = wd.get_dir('centrifuge')
    # if wd.hasDb('Tdb') and (kwargs.get('overwrite',False) == False):
    #     logging.error('Tdb already exists- run with overwrite to overwrite')
    #     sys.exit()

    return Bdb, prod_dir, cent_dir

def check_dependencies(print_out=False):
    '''
    For all possible dependencies, see if you can find them
    '''
    for dep in ['mash', 'nucmer', 'checkm', 'ANIcalculator', 'prodigal', 'centrifuge',
                'nsimscan', 'fastANI']:
        loc, works = find_program(dep)
        works_message = {True:'all good', False:'!!! ERROR !!!'}[works]
        message = '{0:.<40} {1:15} (location = {2})'.format(dep, works_message, loc)
        if print_out:
            print(message)
        else:
            logging.info(message)

def find_program(dep):
    '''
    return location of progrgam, works = True/False (based on calling the help)
    '''
    # find the location of the program
    loc = shutil.which(dep)

    # make sure the help on the program works
    works = False
    if loc != None:
        try:
            o = subprocess.check_output([loc, '-h'],stderr= subprocess.STDOUT)
            works = True
        except:
            pass

    return loc, works

def run_centrifuge(Bdb, prod_dir, cent_dir, **kwargs):
    t = kwargs.get('processors','6')

    cmds = []
    files = []
    for genome in Bdb['genome'].unique():
        genes = "{0}{1}.fna".format(prod_dir, genome)
        cent = "{0}{1}".format(cent_dir, genome)
        if not (os.path.exists("{0}_hits.tsv".format(cent)) and \
                    os.path.exists("{0}_report.tsv".format(cent))):
            cmds.append(gen_centrifuge_cmd(genes,cent,**kwargs))

    if len(cmds) >= 1:
        logging.info('Running Centrifuge')
        for cmd in cmds:
            logging.debug(' '.join(cmd))

        if 'wd' in kwargs:
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
        drep.thread_cmds(cmds, shell=False, logdir=logdir, t=int(t))
        #drep.d_cluster.thread_mash_cmds_status(cmds,t=int(t))

    else:
        logging.info('Past centrifuge runs found- will not re-run')

def gen_read2bin(gene_files):
    r2b = {}
    for f in gene_files:
        genome = os.path.basename(f)[:-4]
        with open(f) as handle:
          for line in handle:
            if line.startswith('>'):
              r2b.setdefault(genome,[]).append(line.strip()[1:].split(' ')[0])
    return r2b

def parse_centrifuge(Bdb, cent_dir, **kwargs):

    Tdb = pd.DataFrame()
    for genome in Bdb['genome'].unique():

        hits = parse_raw_centrifuge("{0}{1}_hits.tsv".format(cent_dir,genome), \
                    "{0}{1}_report.tsv".format(cent_dir,genome))

        if hits.empty:
            logging.debug("No centrifuge hits found for {0}- skipping".format(genome))
            continue

        x = gen_phylo_db(hits)
        x['genome'] = genome
        Tdb = pd.concat([x,Tdb], ignore_index=True)

    # Find the best hits
    g2t = {}
    for genome in Tdb['genome'].unique():
        d = Tdb[Tdb['genome'] == genome]
        taxID = d['tax_ID'][d['tax_confidence'] == d['tax_confidence'].max()].tolist()[0]
        g2t[genome] = taxID
    Tdb['best_hit'] = [True if g2t[g] == t else False for g,t in zip(Tdb['genome'], Tdb['tax_ID'])]

    # Try and add full taxonomy string
    try:
        Tdb['full_tax'] = [lineage_from_taxId(t) if b else False for t, b in zip(\
                Tdb['tax_ID'], Tdb['best_hit'])]
    except:
        logging.info("problem determing full tax string with ete3 - skipping")

    return Tdb

def parse_centrifuge_percent(Bdb, cent_dir, **kwargs):
    min_perc = int(kwargs.get('percent'))
    min_score = kwargs.get('min_score', 250)

    Tdb = pd.DataFrame()
    for genome in Bdb['genome'].unique():
        hits = parse_raw_centrifuge("{0}{1}_hits.tsv".format(cent_dir,genome), \
                    "{0}{1}_report.tsv".format(cent_dir,genome))
        tdb = tdb_from_hits(hits[hits['score'] > min_score], minPerc= int(min_perc))
        tdb['genome'] = genome

        Tdb = pd.concat([Tdb, tdb])
    # Find the best hit

    # THIS IS HARDER BECAUSE YOU HAVE TO PICK THE LOWEST LEVEL OVER the percent
    # # Find the best hits
    # g2t = {}
    # for genome in Tdb['genome'].unique():
    #     d = Tdb[Tdb['genome'] == genome]
    #     taxID = d['tax_ID'][d['tax_confidence'] == d['tax_confidence'].max()].tolist()[0]
    #     g2t[genome] = taxID
    # Tdb['best_hit'] = [True if g2t[g] == t else False for g,t in zip(Tdb['genome'], Tdb['tax_ID'])]
    #
    # # Try and add full taxonomy string
    # try:
    #     Tdb['full_tax'] = [lineage_from_taxId(t) if b else False for t, b in zip(\
    #             Tdb['tax_ID'], Tdb['best_hit'])]
    # except:
    #     logging.info("problem determing full tax string with ete3 - skipping")

    return Tdb

def tdb_from_hits(hits, minPerc= 50, testing=False):
    '''
    Determines the lowest taxonomic level with at least minPerc certainty

    For every hit:
        reconstruct the lineage (kingdom, phylum, class, ect.)
        add a count to every rank in the lineage

    For every rank:
        see if the number of hits matching one taxa at that rank is above the minPerc
        the denominator for this equation is the number of hits that have a phyla rank

    * Note: this is complicated because some lower ranks don't have higher ranks
        For example, species [Eubacterium] rectale (taxID 39491) has no genus
        Also, species [artifical construct] (taxID 32630) has no anything but species

    '''

    from ete3 import NCBITaxa
    ncbi = NCBITaxa()

    Levels = ['superkingdom','phylum','class','order','family','genus','species']

    # generate nested dictionary for levels
    countDic = {}
    for level in Levels:
        countDic[level] = {}

    # fill in nested dictionary
    for t in hits['taxID'].tolist():
        if t == 0:
            continue

        # This try / except thing is trying to catch sporatic errors of:
        # sqlite3.OperationalError: disk I/O error
        try:
            lin = ncbi.get_lineage(t)
            lin2name  = ncbi.get_taxid_translator(lin)
            name2rank = ncbi.get_rank(lin)
        except:
            time.sleep(1)
            lin = ncbi.get_lineage(t)
            lin2name  = ncbi.get_taxid_translator(lin)
            name2rank = ncbi.get_rank(lin)

        for i in lin:
            rank = name2rank[i]
            name = lin2name[i]
            if rank in countDic:
                countDic[rank][i] = countDic[rank].get(i,0) + 1

    # make the table
    total = sum(countDic['phylum'].values())
    table = {'tax_ID':[], 'tax_confidence':[], 'tax_level':[], 'taxonomy':[]}
    count = None

    for level in Levels:
        dic = countDic[level]
        for name in sorted(dic, key=dic.get, reverse= True):
            count = dic[name]
            break

        if count == None:
            table['tax_ID'].append(None)
            table['tax_confidence'].append(0)
            table['tax_level'].append(level)
            table['taxonomy'].append('unk')

        else:
            lin = ncbi.get_lineage(name)
            lin2name  = ncbi.get_taxid_translator(lin)
            name2rank = ncbi.get_rank(lin)
            rank2name = {v: k for k, v in name2rank.items()}
            tax = (lin2name[rank2name[level]])

            table['tax_ID'].append(name)
            table['tax_confidence'].append(((count/total) *100))
            table['tax_level'].append(level)
            table['taxonomy'].append(tax)

        count = None
    tdb = pd.DataFrame(table)

    # find and mark the best hit
    best = tdb['tax_ID'][tdb['tax_confidence'] >= minPerc].tolist()[-1]
    tdb['best_hit'] = [True if i == best else False for i in tdb['tax_ID']]

    # get the full taxonomy for the best hit
    tdb['full_tax'] = [lineage_from_taxId(t) if b else False for t, b in zip(\
            tdb['tax_ID'], tdb['best_hit'])]

    return tdb

def lineage_from_taxId(t):
    from ete3 import NCBITaxa
    ncbi = NCBITaxa()

    Levels = ['superkingdom','phylum','class','order','family','genus','species']
    name = []

    lin = ncbi.get_lineage(t)

    lin2name  = ncbi.get_taxid_translator(lin)
    name2rank = ncbi.get_rank(lin)
    rank2name = {v: k for k, v in name2rank.items()}

    for level in Levels:
        if level in rank2name:
            name.append(lin2name[rank2name[level]])
        else:
            name.append('unk')

    return '|'.join([str(int(t))] + name)

def parse_raw_centrifuge(hits, report):
    tax = pd.read_table(report)
    t2l = tax.set_index('taxID')['taxRank'].to_dict()
    t2n = tax.set_index('taxID')['name'].to_dict()

    hits = pd.read_table(hits)
    hits['gene'] = hits['readID']
    hits['scaffold'] = hits['readID'].map(get_scaff)
    hits['level'] = hits['taxID'].map(t2l)
    hits['name'] = hits['taxID'].map(t2n)
    del hits['readID']

    return hits

def add_taxonomy(Bdb,Tdb):
    g2t = {}
    for genome in Tdb['genome'].unique():
        d = Tdb[(Tdb['genome'] == genome) & (Tdb['best_hit'] == True)]
        if len(d) != 1:
            raise ValueError('Bug')

        tax = d['taxonomy'].tolist()[0]
        g2t[genome] = tax
    Bdb['taxonomy'] = Bdb['genome'].map(g2t)
    return Bdb

def gen_phylo_db(hits):
    Table = {'tax_confidence':[],'taxonomy':[],'tax_level':[],'tax_ID':[]}

    skip = ['uncultured bacterium','Mus musculus', 'Vitis vinifera', 'Homo sapiens']

    levels = ['leaf','subspecies','species','genus','family']
    total_ORFS = len(hits['gene'].unique())
    for level in levels:

        # Restrict to the level in question
        d = hits[hits['level'] == level]

        # Get rid of hits to shitty things
        d = d[~d['name'].isin(skip)]

        # Only get the top hit for each gene
        d = d.sort_values(by='score')
        d = d[~d.duplicated('gene')]

        # Find percentage of top tax
        try:
            top_tax = d['name'].mode()[0]
        except:
            Table['tax_level'].append(level)
            Table['tax_ID'].append(None)
            Table['taxonomy'].append(None)
            Table['tax_confidence'].append(None)
            continue

        x = d[d['name'] == top_tax]
        top_perc = (len(x['gene'].unique()) / total_ORFS) * 100
        tax_ID = x['taxID'].unique()[0]

        Table['tax_level'].append(level)
        Table['tax_ID'].append(tax_ID)
        Table['taxonomy'].append(top_tax)
        Table['tax_confidence'].append(top_perc)

    return pd.DataFrame(Table)

def get_scaff(read):
    return "_".join(read.split('_')[:-1])

def gen_centrifuge_cmd(genes,cent,**kwargs):
    cent_exe = shutil.which('centrifuge')
    if cent_exe == None:
        logging.error("Can't find centrifuge- make sure it's in your system path")
        sys.exit()

    cent_indicies = kwargs.get('cent_index', False)
    if cent_indicies == False:
        logging.error("Can't find centrifuge index- must provide for taxonomy")
        sys.exit()

    cmd = [cent_exe, '-f', '-x', cent_indicies, genes, '-S', "{0}_hits.tsv".format(cent),\
            '-p','1','--report-file',"{0}_report.tsv".format(cent)]

    return cmd



import glob
import logging
import os
import shutil
import sys
import pandas as pd

import drep.d_cluster.external
import drep.d_cluster.compare_utils

def greedy_secondary_clustering(Bdb, Cdb, algorithm, data_folder, **kwargs):
    ndbs = []
    cdbs = []
    c2ret = {}
    for bdb, name in drep.d_cluster.compare_utils.iteratre_clusters(Bdb, Cdb, id='primary_cluster'):
        logging.debug('running cluster {0} with {1} genomes'.format(name, len(bdb)))
        # logging.debug('total memory - {0:.2f} Mbp'.format(int(process.memory_info().rss)/1000000))
        ndb, cdb, ret = drep.d_cluster.compare_utils.compare_genomes(bdb, algorithm, data_folder, **kwargs)

        if len(ndb) == 0:
            logging.error("CRITICAL ERROR WITH PRIMARY CLUSTER {0}; TRYING AGAIN".format(name))
            ndb, cdb, ret = drep.d_cluster.compare_utils.compare_genomes(bdb, algorithm, data_folder, **kwargs)

        if len(ndb) > 0:
            ndb['primary_cluster'] = name
            cdb['primary_cluster'] = name
            cdb['secondary_cluster'] = ["{0}{1}".format(name, y) for y in cdb['secondary_cluster']]
            ndbs.append(ndb)
            cdbs.append(cdb)
            c2ret[name] = ret
        else:
            logging.error("DOUBLE CRITICAL ERROR AGAIN WITH PRIMARY CLUSTER {0}; SKIPPING".format(name))

    Ndb = pd.concat(ndbs).reset_index(drop=True)
    Cdb = pd.concat(cdbs).reset_index(drop=True)
    return Ndb, Cdb, c2ret


def compare_genomes_greedy(bdb, algorithm, data_folder, **kwargs):
    ani_thresh = float(kwargs.get('S_ani', .99))
    cov_thresh = float(kwargs.get('cov_thresh', 0.5))

    assert (ani_thresh <= 1) & (ani_thresh > 0)

    # Set genome order
    odb = order_genomes_for_greedy(bdb, **kwargs)

    # Set up
    cluster = kwargs.get('cluster', '')
    genome_rep_file = os.path.join(data_folder + 'representative_genome_locations.txt')
    genome_reps = []
    rep2cluster = {}
    genome2cluster = {}
    ndbs = []

    if os.path.exists(genome_rep_file):
        os.remove(genome_rep_file)

    # Prepare for greedy clustering
    kwargs = prepare_for_greedy(algorithm, data_folder, **kwargs)

    # Iterate
    j = 1
    for i, row in odb.iterrows():
        if len(genome_reps) == 0:
            make_new_cluster = True

        else:
            ndb = genome_vs_reps(row['location'], genome_reps, genome_rep_file, algorithm, data_folder, **kwargs)
            ndbs.append(ndb)

            cluster_rep = get_cluster_rep(ndb, ani_thresh, cov_thresh)

            if cluster_rep is not False:
                make_new_cluster = False
                assert cluster_rep in rep2cluster, [cluster_rep, rep2cluster]
                genome2cluster[row['genome']] = rep2cluster[cluster_rep]
            else:
                make_new_cluster = True

        if make_new_cluster:
            new_cluster = "{0}_{1}".format(cluster,j)
            j += 1

            genome_reps.append(row['location'])
            rep2cluster[row['genome']] = new_cluster
            genome2cluster[row['genome']] = new_cluster
            with open(genome_rep_file, "a") as myfile:
                myfile.write(row['location'] + '\n')

    if len(ndbs) > 0:
        Ndb = pd.concat(ndbs)

    else:
        # Add self-comparisons if there is only one genome
        Table = {'querry': [], 'reference': [], 'ani': [], 'alignment_coverage': []}
        for g in odb['location'].tolist():
            Table['reference'].append(drep.d_cluster.utils._get_genome_name_from_fasta(g))
            Table['querry'].append(drep.d_cluster.utils._get_genome_name_from_fasta(g))
            Table['ani'].append(1)
            Table['alignment_coverage'].append(1)
        Ndb = pd.DataFrame(Table)

    Cdb, cluster_ret = generate_greedy_cdb(bdb, rep2cluster, genome2cluster, algorithm, ani_thresh, cov_thresh, **kwargs)
    return Ndb, Cdb, cluster_ret

def genome_vs_reps(new_genome, genome_reps, genome_rep_file, algorithm, data_folder, **kwargs):
    if algorithm == 'fastANI':
        # Return Ndb
        return drep.d_cluster.external.fastani_one_vs_many(new_genome, genome_reps, genome_rep_file, data_folder, **kwargs)
    else:
        logging.error("{0} algorithm is not yet supported for greedy clustering; sorry!")
        assert False


def prepare_for_greedy(algorithm, data_folder, **kwargs):
    if algorithm == 'fastANI':
        # Make folders
        if not os.path.exists(data_folder):
            os.makedirs(data_folder)
        tmp_dir = os.path.join(data_folder, 'tmp/')
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)

        # Handle log directory
        if ('wd' in kwargs) and (kwargs.get('debug', False)):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False

        kwargs['tmp_dir'] = tmp_dir
        kwargs['logdir'] = logdir
        kwargs['current_exe'] = drep.get_exe('fastANI')

    return kwargs

def order_genomes_for_greedy(bdb, **kwargs):
    return bdb.sort_values('length', ascending=False)

def get_cluster_rep(ndb, ani_thresh, cov_thresh):
    fdb = ndb[(ndb['ani'] >= ani_thresh) & (ndb['alignment_coverage'] >=  cov_thresh)]
    if len(fdb) > 0:
        return fdb['querry'].iloc[0]
    else:
        return False

def generate_greedy_cdb(bdb, rep2cluster, genome2cluster, algorithm, ani_thresh, c_thresh, **kwargs):
    reps = set(list(rep2cluster.keys()))

    cdb = bdb[['genome']]
    cdb['secondary_cluster'] = cdb['genome'].map(genome2cluster)
    cdb['threshold'] = 1 - ani_thresh
    cdb['cluster_method'] = 'greedy'
    cdb['comparison_algorithm'] = algorithm
    cdb['greedy_representative'] = [g in reps for g in cdb['genome']]

    arguments = {'linkage_method': 'greedy', 'linkage_cutoff': ani_thresh,
                 'comparison_algorithm': algorithm, 'minimum_coverage': c_thresh}
    cluster_ret = [None, pd.DataFrame(), arguments]

    return cdb, cluster_ret




#!/usr/bin/env python3
'''
d_cluster - a subset of dRep

Clusters a list of genomes with both primary and secondary clustering
'''

import pandas as pd
import os
import logging
import random
import string
import numpy as np

import drep
import drep.d_cluster.cluster_utils
import drep.d_cluster.compare_utils
import drep.d_filter
import drep.d_bonus

# from drep.d_cluster import genomeChunk
import drep.d_cluster.controller
#from drep.d_cluster.controller import genomeChunk, genome_hierarchical_clustering, iteratre_clusters, cluster_hierarchical

# This is to make pandas shut up with it's warnings
pd.options.mode.chained_assignment = None

import drep.d_cluster.external
#from drep.d_cluster.external import parse_gani_file, parse_nsim_file, gen_nucmer_cmd, add_avani



def _cluster_Ndb(Ndb, id='primary_cluster', **kwargs):
    '''
    Cluster Ndb on id

    Args:
        Ndb: obvious
        id: thing that defines subclusters in Ndb (defauls = 'primary_cluster')

    Keyword arguments:
        all: passed on to genome_hierarchical_clustering

    Returns:
        list: [Cdb, c2ret(cluster name -> clusering files)]
    '''
    Cdb = pd.DataFrame()
    c2ret = {}
    for name, ndb in Ndb.groupby(id):
        cdb, cluster_ret = drep.d_cluster.cluster_utils.genome_hierarchical_clustering(ndb, cluster=name, **kwargs)
        cdb[id] = name
        Cdb = pd.concat([Cdb,cdb], ignore_index=True)
        c2ret[name] = cluster_ret

    return Cdb, c2ret


def make_linkage_Ndb(Ndb, **kwargs):
    '''
    Filter the Ndb in accordance with the kwargs. Average reciprical ANI values

    Args:
        Ndb: result of secondary clutsering

    Keyword arguments:
        cov_thresh: minimum coverage threshold (default = 0.5)

    Return:
        DataFrame: pivoted DataFrame ready for clustering
    '''
    d = Ndb.copy()
    cov_thresh = float(kwargs.get('cov_thresh',0.5))

    # Remove values without enough coverage
    d.loc[d['alignment_coverage'] <= cov_thresh, 'ani'] = 0

    # Make a linkagedb by averaging values and setting self-compare to 1
    drep.d_cluster.external.add_avani(d)
    d['dist'] = 1 - d['av_ani']
    db = d.pivot("reference", "querry", "dist")

    return db


def estimate_time(comps, alg):
    '''
    Estimate time, in minutes, based on comparison algorithm and number of comparisons

    Args:
        comps: number of genomes comparisons to perform
        alg: algorthm used

    Return:
        float: time to perfom comparison (in minutes)
    '''
    if alg == 'ANIn':
        time = comps * .33
    elif alg == 'gANI':
        time = comps * .1
    elif alg == 'goANI':
        time = comps * .1
    elif alg == 'ANImf':
        time = comps * .5
    elif alg == 'fastANI':
        time = comps * 0.00667
    return time




def _gen_cdb_from_fclust(fclust,names):
    '''
    Make Cdb from the result of scipy.cluster.hierarchy.fcluster

    Args:
        fclust: result of scipy.cluster.hierarchy.fcluster
        names: list(db.columns) of the input dataframe

    Returns:
        DataFrame: Cdb
    '''
    Table={'cluster':[],'genome':[]}
    for i, c in enumerate(fclust):
        Table['cluster'].append(c)
        Table['genome'].append(names[i])

    return pd.DataFrame(Table)


def parse_mash_table(location):
    try:
        iniCols = ['genome1', 'genome2', 'dist', 'p', 'kmers']
        uCols = ['genome1', 'genome2', 'dist']
        dTypes = {'genome1': 'category', 'genome2': 'category', 'dist': np.float32}
        Mdb = pd.read_csv(location, names=iniCols, usecols=uCols, dtype=dTypes, sep='\t')
    except ValueError:
        logging.error("You're using an incompatible version of Mash! I'll try and fix it, but I recommend upgrading Mash to v2.2 or downgrading to v1 if problems occur")
        Mdb = pd.read_csv(location, sep='\t')

    Mdb['genome1'] = Mdb['genome1'].apply(_get_genome_name_from_fasta).astype('category')
    Mdb['genome2'] = Mdb['genome2'].apply(_get_genome_name_from_fasta).astype('category')
    Mdb['similarity'] = 1 - Mdb['dist']

    return Mdb


def merge_genome_chunks(mash_exe, genome_chunks, sketch_folder, MASH_folder):
    all_file = os.path.join(MASH_folder, 'ALL.msh')
    cmd = [mash_exe, 'paste', all_file] + [gc.all_file for gc in genome_chunks]

    new_locs = []
    new_names = []
    for gc in genome_chunks:
        new_locs += gc.genome_locations
        new_names += gc.genome_names

    new_gc = drep.d_cluster.compare_utils.genomeChunk(new_locs, 'all', sketch_folder, new_names, no_create=True)
    new_gc.all_file = all_file

    return cmd, new_gc


# def all_vs_all_MASH(Bdb, data_folder, **kwargs):
#     """
#     Run MASH pairwise within all samples in Bdb
#
#     Args:
#         Bdb: dataframe with genome, location
#         data_folder: location to store output files
#
#     Keyword Args:
#         MASH_sketch: size of mash sketches
#         dry: dont actually run anything
#         processors: number of processors to multithread with
#         mash_exe: location of mash excutible (will try and find with shutil if not provided)
#         groupSize: max number of mash sketches to hold in each folder
#         debug: if True, log all of the commands
#         wd: if you want to log commands, you also need the wd
#     """
#
#     MASH_s = kwargs.get('MASH_sketch',1000)
#     dry = kwargs.get('dry',False)
#     # overwrite = kwargs.get('overwrite', False)
#     p = kwargs.get('processors',6)
#     groupSize = kwargs.get('groupSize', 1000)
#
#     # set up logdir
#     if ('wd' in kwargs) and (kwargs.get('debug', False) == True):
#         logdir = kwargs.get('wd').get_dir('cmd_logs')
#     else:
#         logdir = False
#
#     # Find mash
#     mash_exe = kwargs.get('exe_loc', None)
#     if mash_exe == None:
#         mash_exe = drep.get_exe('mash')
#
#     # Set up folders
#     MASH_folder = os.path.join(data_folder, 'MASH_files/')
#     if not os.path.exists(MASH_folder):
#         os.makedirs(MASH_folder)
#
#     sketch_folder = os.path.join(MASH_folder, 'sketches/')
#     if not os.path.exists(sketch_folder):
#         os.makedirs(sketch_folder)
#
#     # Make chunks
#     l2g = Bdb.set_index('location')['genome'].to_dict()
#     locations = list(Bdb['location'].unique())
#     chunks = [locations[x:x+groupSize] for x in range(0, len(locations), groupSize)]
#
#     # Make the MASH sketches
#     cmds = []
#     chunk_folders = []
#     for i, chunk in enumerate(chunks):
#         chunk_folder = os.path.join(sketch_folder, "chunk_{0}".format(i))
#         chunk_folders.append(chunk_folder)
#         if not os.path.exists(chunk_folder):
#             os.makedirs(chunk_folder)
#         for fasta in chunk:
#             genome = l2g[fasta]
#             file = os.path.join(chunk_folder, genome)
#             if not os.path.isfile(file + '.msh'):
#                 cmd = [mash_exe, 'sketch', fasta, '-s', str(MASH_s), '-o',
#                     file]
#                 cmds.append(cmd)
#
#     if not dry:
#         if len(cmds) > 0:
#             drep.thread_cmds(cmds, logdir=logdir, t=int(p))
#
#     # Combine MASH sketches within chunk
#     cmds = []
#     alls = []
#     for chunk_folder in chunk_folders:
#         all_file = os.path.join(chunk_folder, 'chunk_all.msh')
#         cmd = [mash_exe, 'paste', all_file] \
#                 + glob.glob(os.path.join(chunk_folder, '*'))
#         cmds.append(cmd)
#         alls.append(all_file)
#     if not dry:
#         if len(cmds) > 0:
#             drep.thread_cmds(cmds, logdir=logdir, t=int(p))
#
#     # Combine MASH sketches of all chunks
#     all_file = os.path.join(MASH_folder, 'ALL.msh')
#     cmd = [mash_exe, 'paste', all_file] + alls
#     drep.run_cmd(cmd, dry, shell=False, logdir=logdir)
#
#     # Calculate distances
#     cmd = [mash_exe, 'dist','-p', str(p), all_file, all_file, '>', MASH_folder
#             + 'MASH_table.tsv']
#     cmd = ' '.join(cmd)
#     drep.run_cmd(cmd, dry, shell=True, logdir=logdir)
#
#     # Make Mdb based on all genomes in the MASH folder
#     file = MASH_folder + 'MASH_table.tsv'
#
#     iniCols = ['genome1','genome2','dist','p','kmers']
#     uCols = ['genome1','genome2','dist']
#     dTypes = {'genome1':'category', 'genome2':'category', 'dist':np.float32}
#     Mdb = pd.read_csv(file, names=iniCols, usecols=uCols, dtype=dTypes, sep='\t')
#     Mdb['genome1'] = Mdb['genome1'].apply(_get_genome_name_from_fasta).astype('category')
#     Mdb['genome2'] = Mdb['genome2'].apply(_get_genome_name_from_fasta).astype('category')
#     Mdb['similarity'] = 1 - Mdb['dist']
#
#     # Filter out those genomes that are in the MASH folder but shouldn't be in Mdb
#     genomes = Bdb['genome'].unique()
#     Mdb = Mdb[Mdb['genome1'].isin(genomes)]
#     Mdb = Mdb[Mdb['genome2'].isin(genomes)]
#
#     # Reorder categories to be correct
#     for g in ['genome1', 'genome2']:
#         Mdb[g] = Mdb[g].cat.remove_unused_categories()
#         Mdb[g] = Mdb[g].cat.reorder_categories(sorted((Mdb[g].unique())), ordered=True)
#
#     return Mdb

def _get_genome_name_from_fasta(fasta):
    '''
    Just do a os.path.basename command

    Args:
        fasta: location of .fasta file

    Returns:
        string: basename of .fasta files
    '''
    return str(os.path.basename(fasta))

def parse_delta(filename):
    '''
    Parse a .delta file from nucmer

    Args:
        filename: location of .delta file

    Returns:
        list: [alignment_length, similarity_errors]
    '''

    aln_length, sim_errors = 0, 0
    for line in [l.strip().split() for l in open(filename, 'r').readlines()]:
        if line[0] == 'NUCMER' or line[0].startswith('>'):  # Skip headers
            continue
        # We only process lines with seven columns:
        if len(line) == 7:
            aln_length += abs(int(line[1]) - int(line[0]))
            sim_errors += int(line[4])
    return aln_length, sim_errors

def gen_gANI_cmd(file, g1, g2, exe):
    '''
    Generate a command to run gANI (ANIcalculator)

    Will return [] if g1 == g2

    Args:
        file: output file name
        g1: location of the genes of genome1
        g2: location of the genes of genome2
        exe: location of gANI executible

    Returns:
        list: command to run
    '''
    # Don't do self-comparions
    if g1 == g2:
        return []

    else:
        cmd = [exe,'-genome1fna',g1,'-genome2fna',g2,'-outfile',file,'-outdir',file + 'TEMP']
    return cmd

def gen_goANI_cmd(file, g1, g2, exe):
    '''
    Generate a command to run goANI

    Will return [] if g1 == g2

    Args:
        file: output file name
        g1: location of the genes of genome1
        g2: location of the genes of genome2
        exe: location of nsimscan executible

    Returns:
        list: command to run
    '''
    # Don't do self-comparions
    if g1 == g2:
        return []

    else:
        cmd = [exe,'--om','TABX', g1, g2, file]
    return cmd

def process_deltafiles(deltafiles, org_lengths, logger=None, **kwargs):
    '''
    Parse a list of delta files into a pandas dataframe

    Files NEED to be named in the format genome1_vs_genome2.delta, or genome1_vs_genome2.filtered.delta

    Args:
        deltafiles: list of .delta files
        org_lengths: dictionary of genome to genome length

    Keyword Arguments:
        coverage_method: default is larger
        logger: if not None, will log 0 errors

    Returns:
        DataFrame: information about the alignments
    '''

    Table = {'querry':[],'reference':[],'alignment_length':[],'similarity_errors':[],
            'ref_coverage':[],'querry_coverage':[],'ani':[], 'reference_length':[],
            'querry_length':[],'alignment_coverage':[]}

    coverage_method = kwargs.get('coverage_method', 'larger')

    for deltafile in deltafiles:
        qname, sname = os.path.splitext(os.path.split(deltafile)[-1])[0].split('_vs_')
        # Handle naming of "filtered"
        if sname.endswith('.delta'):
            sname = sname[:-6]

        tot_length, tot_sim_error = parse_delta(deltafile)
        if tot_length == 0 and logger is not None:
            logging.info("Total alignment length reported in " +
                               "%s is zero!" % deltafile)
        query_cover = float(tot_length) / org_lengths[qname]
        sbjct_cover = float(tot_length) / org_lengths[sname]

        try:
            perc_id = 1 - float(tot_sim_error) / tot_length
        except ZeroDivisionError:
            #print("Alignment between {0} and {1} has 0 alignment!".format(qname,sname))
            perc_id = 0  # set arbitrary value of zero identity


        Table['querry'].append(qname)
        Table['querry_length'].append(org_lengths[qname])
        Table['reference'].append(sname)
        Table['reference_length'].append(org_lengths[sname])
        Table['alignment_length'].append(tot_length)
        Table['similarity_errors'].append(tot_sim_error)
        Table['ani'].append(perc_id)
        Table['ref_coverage'].append(sbjct_cover)
        Table['querry_coverage'].append(query_cover)

        if coverage_method == 'total':
            Table['alignment_coverage'].append((tot_length * 2)/(org_lengths[qname]\
                                                             + org_lengths[sname]))
        elif coverage_method == 'larger':
            Table['alignment_coverage'].append(max((tot_length/org_lengths[qname]),\
                (tot_length/org_lengths[sname])))

    dTypes = {'querry':'category','querry_length':np.int64, 'reference':'category',\
            'reference_length':np.int64, 'alignment_length':np.int64, 'similarity_errors':np.int64,\
            'ani':np.float32, 'ref_coverage':np.float32, 'querry_coverage':np.float32}
    df = pd.DataFrame(Table)
    for c, t in dTypes.items():
        df[c] = df[c].astype(t)
    return df

def process_goani_files(files):
    '''
    From a list of nsimscan files, return a parsed DataFrame like normal,
    and a special one for calculating dn/ds

    Args:
        list: files

    Returns:
        DataFrame: Ndb
    '''
    Table = {'querry':[],'reference':[],'ani':[],'alignment_coverage':[]}
    for file in files:
        results = drep.d_cluster.external.parse_nsim_file(file)
        results['reference'] = os.path.basename(file).split('_vs_')[0]
        results['querry'] = os.path.basename(file).split('_vs_')[1][:-4]

        # Add forward results
        Table['reference'].append(results['reference'])
        Table['querry'].append(results['querry'])
        Table['ani'].append(float(results['ani'])/100)
        Table['alignment_coverage'].append(results['af'])
        #Table['alignment_coverage'].append(results['rq_coverage'])

        # Add reverse results [they're the same]
        # Table['reference'].append(results['querry'])
        # Table['querry'].append(results['reference'])
        # Table['ani'].append(float(results['ani'])/100)
        # Table['alignment_coverage'].append(results['af'])
        #Table['alignment_coverage'].append(results['qr_coverage'])

    # Add self comparisons
    for g in set(Table['reference']):
        Table['reference'].append(g)
        Table['querry'].append(g)
        Table['ani'].append(1)
        Table['alignment_coverage'].append(1)

    Gdb = pd.DataFrame(Table)
    return Gdb

def process_gani_files(files):
    '''
    From a list of gANI output files, return a parsed DataFrame

    Args:
        list: files

    Returns:
        DataFrame: Ndb
    '''
    Table = {'querry':[],'reference':[],'ani':[],'alignment_coverage':[]}
    for file in files:
        results = drep.d_cluster.external.parse_gani_file(file)

        # calculate coverage using the "larger" method
        cov = max(results['rq_coverage'], results['qr_coverage'])

        # Add forward results
        Table['reference'].append(results['reference'])
        Table['querry'].append(results['querry'])
        Table['ani'].append(float(results['rq_ani'])/100)
        Table['alignment_coverage'].append(cov)
        #Table['alignment_coverage'].append(results['rq_coverage'])

        # Add reverse results [they're the same]
        Table['reference'].append(results['querry'])
        Table['querry'].append(results['reference'])
        Table['ani'].append(float(results['qr_ani'])/100)
        Table['alignment_coverage'].append(cov)
        #Table['alignment_coverage'].append(results['qr_coverage'])

    # Add self comparisons
    for g in set(Table['reference']):
        Table['reference'].append(g)
        Table['querry'].append(g)
        Table['ani'].append(1)
        Table['alignment_coverage'].append(1)

    Gdb = pd.DataFrame(Table)
    return Gdb

def run_pairwise_ANIn(genome_list, ANIn_folder, **kwargs):
    '''
    Given a list of genomes and an output folder, compare all genomes using ANImf

    Args:
        genome_list: list of locations of genome files
        ANIn_folder: folder to store the output of comparison

    Keyword arguments:
        processors: threads to use
        debug: if true save extra output
        wd: needed if debug is True
    '''
    p = kwargs.get('processors',6)
    genomes = genome_list

    # Make folder
    if not os.path.exists(ANIn_folder):
        os.makedirs(ANIn_folder)

    # Gen commands
    cmds = []
    files = []
    for g1 in genomes:
        # Make it so each reference is it's own folder, to spread out .delta files
        cur_folder = os.path.join(ANIn_folder, _get_genome_name_from_fasta(g1))
        if not os.path.exists(cur_folder):
            os.makedirs(cur_folder)

        for g2 in genomes:
            file_name = "{0}{1}_vs_{2}".format(ANIn_folder, \
                        _get_genome_name_from_fasta(g1),\
                        _get_genome_name_from_fasta(g2))
            files.append(file_name)

            # If the file doesn't already exist, add it to what needs to be run
            if not os.path.isfile(file_name + '.delta'):
                cmds.append(drep.d_cluster.external.gen_nucmer_cmd(file_name, g1, g2))

    # Run commands
    if len(cmds) > 0:
        for c in cmds:
            logging.debug(' '.join(c))

        if ('wd' in kwargs) and (kwargs.get('debug', False)):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
        drep.thread_cmds(cmds, shell=False, logdir=logdir, t=int(p))

    # Make dictionary of genome lengths
    org_lengths = {}
    for genome in genomes:
        org_lengths[_get_genome_name_from_fasta(genome)] = \
            drep.d_filter.calc_fasta_length(genome)



    deltafiles = ["{0}.delta".format(file) for file in files]
    df = process_deltafiles(deltafiles, org_lengths, **kwargs)

    return df

def load_genomes(genome_list):
    '''
    Takes a list of genome locations, returns a pandas dataframe with the
    locations and the basename of the genomes

    Args:
        genome_list: list of genome locations

    Returns:
        DataFrame: pandas dataframe with columns ['genome', 'location']
    '''
    assert type(genome_list) == type(list())

    # Load from a text file
    if len(genome_list) == 1:
        logging.info('Loading genomes from a list')
        try:
            Table = {'genome':[],'location':[]}
            with open(genome_list[0], 'r') as o:
                for line in o.readlines():
                    genome = line.strip()
                    assert os.path.isfile(genome), "{0} is not a file".format(genome)
                    Table['genome'].append(os.path.basename(genome))
                    Table['location'].append(os.path.abspath(genome))
            Bdb = pd.DataFrame(Table)
            return Bdb
        except:
            logging.info('Nevermind! Ill try loading as a genome now')

    Table = {'genome':[], 'location':[]}
    for genome in genome_list:
        assert os.path.isfile(genome), "{0} is not a file".format(genome)
        Table['genome'].append(os.path.basename(genome))
        Table['location'].append(os.path.abspath(genome))
    Bdb = pd.DataFrame(Table)
    return Bdb


def _gen_nomani_cdb(Cdb, **kwargs):
    '''
    Make a Cdb that looks like the Cdb from secondary clustering

    Args:
        Cdb: Result from Mash clustering

    Returns:
        DataFrame: Cdb
    '''
    c_method = kwargs.get('clusterAlg','single')
    threshold = 1 - kwargs.get('P_ani',.9)

    # Make Cdb look like Cdb
    cdb = Cdb.copy()
    cdb['secondary_cluster'] = ["{0}_0".format(i) for i in cdb['primary_cluster']]
    cdb.rename(columns={'primary_cluster': 'primary_cluster'}, inplace=True)
    cdb['threshold'] = threshold
    cdb['cluster_method'] = c_method
    cdb['comparison_algorithm'] = 'MASH'

    return cdb

def _print_time_estimate(Bdb, Cdb, algorithm, cores):
    '''
    Print an estimate of how long genome comaprisons will take

    Args:
        Bdb: DataFrame with [genome, location]
        Cdb: Clustering DataFrame
        algorthm: algorithm to estimate time with
    '''
    if algorithm != 'fastANI':
        comps = 0
        for bdb, name in drep.d_cluster.cluster_utils.iteratre_clusters(Bdb, Cdb):
            g = len(bdb['genome'].unique())
            comps += (g * g)

        time = estimate_time(comps, algorithm)
        time = time / int(cores)

    else:
        time = 0
        comps = 0
        for bdb, name in drep.d_cluster.cluster_utils.iteratre_clusters(Bdb, Cdb):
            g = len(bdb['genome'].unique())
            time += predict_fastani_time(g, cores)
            comps += (g * g)
        # Convert to minutes
        time = time / 60

    logging.info("Running {0} {1} comparisons- should take ~ {2:.1f} min".format(\
            comps, algorithm, time))


def predict_fastani_time(genomes, cores=6):
    '''
    Forumula is f(x) = a*x**2 + b*x

    http://localhost:8888/doc/tree/Other/OneOffs/dRep_performance_fastANI_timing_1.ipynb
    '''
    a = 0.018969442838018713
    b = 14.604906958328073

    genomes = int(genomes)
    time_6cores = a * genomes ** 2 + b * genomes

    time = time_6cores * (6 / cores)

    return time

def _randomString(stringLength=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))



#!/usr/bin/env python3
"""
This module provides access to the workDirectory

The directory layout::

    workDirectory
    ./data
    ...../MASH_files/
    ...../ANIn_files/
    ...../gANI_files/
    ...../Clustering_files/
    ...../checkM/
    ........./genomes/
    ........./checkM_outdir/
    ...../prodigal/
    ./figures
    ./data_tables
    ...../Bdb.csv  # Sequence locations and filenames
    ...../Mdb.csv  # Raw results of MASH comparisons
    ...../Ndb.csv  # Raw results of ANIn comparisons
    ...../Cdb.csv  # Genomes and cluster designations
    ...../Chdb.csv # CheckM results for Bdb
    ...../Sdb.csv  # Scoring information
    ...../Wdb.csv  # Winning genomes
    ./dereplicated_genomes
    ./log
    ...../logger.log
    ...../cluster_arguments.json

"""

import os
import logging
import pandas as pd
import pickle
import json
import sys
import shutil
import glob
import numpy as np

import drep

class WorkDirectory(object):
    '''
    Object to interact with the workDirectory

    Args:
        location (str): location to make the workDirectory


    '''
    firstLevels = ['data','figures','data_tables','dereplicated_genomes','log']

    def __init__(self, location):
        self.location = os.path.abspath(location)
        self.data_tables = {}
        self.clusters = {}
        self.arguments = {}
        self.overwrite = True
        self.name = None

        self.make_fileStructure()
        self.load_cached()

    def __str__(self):
        string = "Located: {0}\nDatatables: {1}\nCluster files: {2}\nArguments: {3}".format(\
                    self.location,list(self.data_tables.keys()),list(self.clusters.keys()),\
                    list(self.arguments.keys()))

        return string

    def make_fileStructure(self):
        '''
        Make the top level file structure
        '''
        location = self.location

        if not os.path.exists(location):
            os.makedirs(location)

        for l in WorkDirectory.firstLevels:
            loc = location + '/' + l
            if not os.path.exists(loc):
                os.makedirs(loc)

    def load_cached(self):
        '''
        The wrapper to load everything it has into attributes
        '''
        # Import data_tables
        loc = self.get_dir('data_tables')
        self.import_data_tables(loc)

        # Import pickles
        loc = self.location + '/data/Clustering_files/'
        if not os.path.exists(loc):
            os.makedirs(loc)
        self.import_clusters(loc)

        # Import arguments
        loc = self.location + '/log/'
        if not os.path.exists(loc):
            os.makedirs(loc)
        self.import_arguments(loc)

    def import_data_tables(self, loc):
        '''
        Given the location of the datatables, load them
        '''
        tables = [os.path.join(loc, t) for t in os.listdir(loc) if \
                os.path.isfile(os.path.join(loc, t))]
        for t in tables:
            self.data_tables[os.path.basename(t).replace('.csv','').\
                replace('.pickle', '')] = t

    def import_clusters(self,loc):
        '''
        Given the location of the cluster files, load them
        '''
        pickles = [os.path.join(loc, t) for t in os.listdir(loc) if os.path.isfile(os.path.join(loc, t))]
        for p in pickles:
            assert p.endswith('.pickle'), "{0} is incorrectly in the data/Clustering_files folder".format(p)

            f = open(p,'rb')
            try:
                linkage = pickle.load(f)
                db = pickle.load(f)
                args = pickle.load(f)
                name = os.path.basename(p).replace('.pickle','')
                self.clusters[name] = {'linkage':linkage,'db':db,'arguments':args}
            except:
                print("{0} is an imporperly made pickle- skipping import".format(p))

    def import_arguments(self,loc):
        '''
        Given the location of the log directory, load it
        '''
        args = [os.path.join(loc, t) for t in os.listdir(loc) if os.path.isfile(os.path.join(loc, t))]
        for j in args:
            if j.endswith('_arguments.json'):
                with open(j) as data_file:
                    data = json.load(data_file)
                self.arguments[os.path.basename(j).replace('_arguments.json','')] = data

    def hasDb(self,db):
        '''
        If db is in the data_tables, return True
        '''
        return (db in self.data_tables)

    def get_cluster(self, name):
        '''
        Get the cluster passed in

        Args:
            name: name of the cluster

        Returns:
            cluster
        '''
        # If the whole cluster name is passed in
        if name in self.clusters:
            return self.clusters[name]

        else:
            n = 'secondary_linkage_cluster_{0}'.format(name)
            if n in self.clusters:
                return self.clusters[n]

        print("Could not find cluster {0} - could it be a singleton?".format(name))
        print("The clusters I have are: {0}".format(list(self.clusters.keys())))
        print("Quitting")
        sys.exit()

    def get_primary_linkage(self):
        '''
        Get the primary linkage cluster
        '''
        return self.clusters['primary_linkage']

    def store_db(self,db,name,overwrite=None):
        '''
        Store a dataframe in the workDirectory

        Will make a physical copy in the datatables folder

        Args:
            db: pandas dataframe to store
            name: name to store it under (will add .csv automatically)
            overwrite: if True, overwrite if DataFrame with same name already exists
        '''
        loc = self.get_dir('data_tables')

        if overwrite == None:
            overwrite = self.overwrite

        if os.path.isfile(loc + name + '.csv'):
            assert overwrite == True, "data_table {0} already exists".format(name)

        if name == 'Mdb':
            floc = loc + name + '.csv'
            db.to_csv(floc, index=False)
            self.data_tables[name] = floc

        else:
            floc = loc + name + '.csv'
            db.to_csv(floc, index=False)
            self.data_tables[name] = floc

    def get_db(self, name, return_none=True, forPlotting=False):
        '''
        Get database from self.data_tables

        Args:
            name: name of dataframe
            return_none: if True will return None if database not found; otherwise assert False
            forPlotting: if True don't do fancy dType loading; it messes with order of names for dendrograms
        '''
        if name in self.data_tables:
            if name == 'Mdb':
                if forPlotting:
                    return pd.read_csv(self.data_tables[name])
                else:
                    dTypes={'genome1':'category', 'genome2':'category', 'dist':np.float32,\
                        'similarity':np.float32}
                    return pd.read_csv(self.data_tables[name], dtype=dTypes)

            else:
                return pd.read_csv(self.data_tables[name])
        else:
            if return_none:
                return None
            else:
                assert False, "Datatable {0} is not in the work directory {1}".format(\
                                name, self.location)

    def get_dir(self, dir):
        '''
        Get the location of one of the named directory types

        Args:
            dir: Name of directory to find

        Returns:
            string: Location of requested directory
        '''
        d = None
        if dir == 'data_tables':
            d = self.location + '/data_tables/'
        if dir == 'prodigal':
            d = self.location + '/data/prodigal/'
        elif dir == 'centrifuge':
            d = self.location + '/data/centrifuge/'
        elif dir == 'ESOM':
            d = self.location + '/data/ESOM/'
        elif dir == 'log':
            d = self.location + '/log/'
        elif dir == 'cmd_logs':
            d = self.location + '/log/cmd_logs/'
        elif dir == 'MASH':
            d = self.location + '/data/MASH_files/'
        elif dir == 'checkM':
            d = self.location + '/data/checkM/'
        elif dir == 'data':
            d = self.location + '/data/'
        elif dir == 'clustering':
            d = self.location + '/data/Clustering_files/'
        elif dir == 'dereplicated_genomes':
            d = self.location + '/dereplicated_genomes/'
        elif dir == 'figures':
            d = self.location + '/figures/'

        if d == None:
            assert False, "{0} is not a directory I know about".format(dir)

        if not os.path.exists(d):
            os.makedirs(d)

        return d

    def get_loc(self, what):
        '''
        Get the location of Things

        Args:
            what: string of what to get the location of

        Returns:
            string: location of what
        '''
        if what == 'log':
            return self.location + '/log/logger.log'

    def store_special(self, name, thing):
        '''
        Store special items in the work directory

        Args:
            name: what to store
            thing: actual thing to store
        '''
        if name == 'primary_linkage':
            assert len(thing) == 3
            store_loc = self.get_dir('clustering')
            logging.debug('Saving primary_linkage pickle to {0}'.format(store_loc))
            with open(store_loc + 'primary_linkage.pickle', 'wb') as handle:
                pickle.dump(thing[0], handle, protocol=4)
                pickle.dump(thing[1], handle, protocol=4)
                pickle.dump(thing[2], handle, protocol=4)

        elif name == 'secondary_linkages':
            assert type(thing) == dict
            store_loc = self.get_dir('clustering')
            for name, cluster_ret in thing.items():
                #print(name)
                if len(cluster_ret) == 0:
                    continue
                assert len(cluster_ret) == 3

                pickle_name = "secondary_linkage_cluster_{0}.pickle".format(name)
                logging.debug('Saving secondary_linkage pickle {1} to {0}'.format(pickle_name,\
                                                                    store_loc))
                with open(store_loc + pickle_name, 'wb') as handle:
                    pickle.dump(cluster_ret[0],handle, protocol=4)
                    pickle.dump(cluster_ret[1],handle, protocol=4)
                    pickle.dump(cluster_ret[2],handle, protocol=4)

        elif name == 'dereplicated_genomes':
            output_folder = self.get_dir('dereplicated_genomes')
            if os.path.exists(output_folder):
                #logging.debug("{0} already exists: removing and remaking".format(output_folder))
                shutil.rmtree(output_folder)
            os.makedirs(output_folder)

            for loc in thing:
                genome = os.path.basename(loc)
                shutil.copy2(loc, "{0}{1}".format(output_folder,genome))
            #logging.info("Done! Dereplicated genomes saved at {0}".format(output_folder))

        elif name == 'cluster_log':
            cluster_log = os.path.join(self.get_dir('log') + 'cluster_arguments.json')
            with open(cluster_log, 'w') as fp:
                json.dump(thing, fp)
            fp.close()

    def _wipe_secondary_clusters(self):
        '''
        Wipe any and all secondary clusters present in the workDirectory
        '''
        # Clear out clustering folder
        c_folder = self.get_dir('clustering')
        for fn in glob.glob(c_folder + 'secondary_linkage_cluster*'):
            os.remove(fn)



import logging
import shutil

import pandas as pd

import drep
import drep.d_cluster.external
import drep.d_cluster.utils
import drep.d_cluster.compare_utils

class GenomeClusterController(object):
    """
    Handle the logic of comparing and clustering genomes
    """
    def __init__(self, wd, **kwargs):
        self.wd = drep.WorkDirectory.WorkDirectory(wd)
        self.kwargs = kwargs

        # Handle special kwargs
        self.debug = kwargs.get('debug', False)

    def main(self, store_output=True):
        """
        Main entrypoint for the dRep cluster operation

        Formerly called d_cluster_wrapper
        """
        # Get the arguments
        self.parse_cluster_arguments()

        # Run primary clustering
        self.run_primary_clustering()

        # Run secondary clustering
        self.run_secondary_clustering()

        # Save the output
        if store_output:
            self.store_output()
        else:
            self.return_output()

    def parse_cluster_arguments(self):
        """
        Load the genomes and store Bdb in the wd
        """
        # Make sure you have the required program installed
        loc = shutil.which('mash')
        if loc is None:
            logging.error('Cannot locate the program {0}- make sure its in the system path' \
                          .format('mash'))

        # If genomes are provided, load them
        if self.kwargs.get('genomes', None) is not None:
            assert self.wd.hasDb("Bdb") == False, \
                "Don't provide new genomes- you already have them in the work directory"
            Bdb = drep.d_cluster.utils.load_genomes(self.kwargs['genomes'])

        # If genomes are not provided, don't load them
        if self.kwargs.get('genomes', None) is None:
            assert self.wd.hasDb("Bdb") != False, \
                "Must either provide a genome list, or run the 'filter' operation with the same work directory"
            Bdb = self.wd.get_db('Bdb')

        # Make sure people weren't dumb with their cutoffs
        for v in ['P_ani', 'S_ani']:
            if self.kwargs.get(v) > 1:
                logging.warning("{0} is set to {1}- this should be \
                            between 0-1, not 1-100".format(v, self.kwargs.get(v)))

        # Load length and N50 if you need it
        if self.kwargs.get('multiround_primary_clustering', False) | (self.kwargs.get('greedy_secondary_clustering', False)):
            Bdb = drep.d_filter._add_lengthN50(Bdb, Bdb)

        # Store the genomes
        self.Bdb = Bdb

    def run_primary_clustering(self):
        """
        Run primary clustering and end with an Mdb and an MCdb in this object
        """
        logging.info("Running primary clustering")
        cached = (self.debug and self.wd.hasDb('Mdb') and self.wd.hasDb('CdbF'))

        if self.kwargs.get('SkipMash', False):
            logging.info("Nevermind! Skipping Mash")
            # Make a "Cdb" where all genomes are in the same cluster
            Cdb = drep.d_cluster.external._gen_nomash_cdb(self.Bdb)
            # Make a blank "Mdb" for storage anyways
            Mdb = pd.DataFrame({'Blank': []})

        elif cached:
            logging.info('Nevermind! Loading cached primary clustering')
            Mdb = self.wd.get_db('Mdb')
            Cdb = self.wd.get_db('CdbF')
            logging.info('2. Primary clustering cache loaded')

        else:
            logging.info("Running pair-wise MASH clustering")
            Mdb, Cdb, cluster_ret = drep.d_cluster.compare_utils.all_vs_all_MASH(self.Bdb, self.wd.get_dir('MASH'), **self.kwargs)

            if self.debug:
                logging.debug("Debug mode on - saving Mdb ASAP")
                self.wd.store_db(Mdb, 'Mdb')

                logging.debug("Debug mode on - saving CdbF ASAP")
                self.wd.store_db(Cdb, 'CdbF')

            # Store the primary clustering results
            self.wd.store_special('primary_linkage', cluster_ret)

        logging.info("{0} primary clusters made".format(len(Cdb['primary_cluster'].unique())))
        self.Mdb = Mdb
        self.MCdb = Cdb

    def run_secondary_clustering(self):
        logging.info("Running secondary clustering")

        # Get the arguments
        algorithm = self.kwargs.get('S_algorithm', 'ANImf')
        cached = (self.debug and self.wd.hasDb('Ndb'))
        p = self.kwargs.get('processors', 6)
        self.deal_with_nucmer_presets()

        # Wipe any old secondary clusters
        self.wd._wipe_secondary_clusters()

        if not self.kwargs.get('SkipSecondary', False):
            if cached:
                logging.info('3. Loading cached secondary clustering')
                Ndb = self.wd.get_db('Ndb')
                # Get rid of broken ones
                Ndb = Ndb.dropna(subset=['reference'])

                logging.info('3. Secondary clustering cache loaded')

            # Run comparisons, make Ndb
            else:
                drep.d_cluster.utils._print_time_estimate(self.Bdb, self.MCdb, algorithm, p)
                Ndb, Cdb, c2ret = drep.d_cluster.compare_utils.secondary_clustering(self.Bdb, self.MCdb, algorithm, self.wd.get_dir('data'), wd=self.wd, **self.kwargs)
                if self.debug:
                    logging.debug("Debug mode on - saving Ndb ASAP")
                    self.wd.store_db(Ndb, 'Ndb')

            # Store the secondary clustering results
            self.wd.store_special('secondary_linkages', c2ret)

        else:
            logging.info("3. Nevermind! Skipping secondary clustering")
            Cdb = drep.d_cluster.utils._gen_nomani_cdb(self.MCdb, data_folder=self.wd.get_dir('data'), **self.kwargs)
            Ndb = pd.DataFrame({'Blank': []})

        logging.info(
            "Step 4. Return output")

        self.Cdb = Cdb
        self.Ndb = Ndb

    def store_output(self):
        logging.debug("Main program run complete- saving output")
        self.wd.store_db(self.Cdb, 'Cdb')
        self.wd.store_db(self.Mdb, 'Mdb')
        self.wd.store_db(self.Ndb, 'Ndb')
        if not self.wd.hasDb('Bdb'):
            self.wd.store_db(self.Bdb, 'Bdb')

        # Log arguments
        self.wd.store_special('cluster_log', self.kwargs)

    def return_output(self):
        return self.Cdb, self.Mdb, self.Ndb

    def deal_with_nucmer_presets(self):
        if self.kwargs.get('n_preset', None) != None:
            self.kwargs['n_c'], self.kwargs['n_maxgap'], self.kwargs['n_noextend'], self.kwargs['n_method'] \
                = drep.d_cluster.external._nucmer_preset(self.kwargs['n_PRESET'])

def d_cluster_wrapper(workDirectory, **kwargs):
    GenomeClusterController(workDirectory, **kwargs).main()





#!/usr/bin/env python

'''
dRep- parse command-line arguemnts
'''

__author__ = "Matt Olm"
__license__ = "MIT"
__email__ = "mattolm@gmail.com"
__status__ = "Development"

import argparse
import os
import sys

import drep
from drep.controller import Controller


def version():
    versionFile = open(os.path.join(drep.__path__[0], 'VERSION'))
    return versionFile.read().strip()


VERSION = version()

"""
########################################
#    Argument Parsing                  #
########################################
"""


class SmartFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)


def printHelp():
    print('')
    print('                ...::: dRep v' + VERSION + ' :::...''')
    print('''\

  Matt Olm. MIT License. Banfield Lab, UC Berkeley. 2017 (last updated 2022)

  See https://drep.readthedocs.io/en/latest/index.html for documentation
  Choose one of the operations below for more detailed help. 
  
  Example: dRep dereplicate -h

  Commands:
    compare            -> Compare and cluster a set of genomes
    dereplicate        -> De-replicate a set of genomes
    check_dependencies -> Check which dependencies are properly installed
    ''')


def parse_args(args):
    parser = argparse.ArgumentParser(formatter_class=SmartFormatter)
    subparsers = parser.add_subparsers(help='Desired operation', dest='operation')

    # Make a parent parser for all of the subparsers
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument("work_directory", help="R|Directory where data and output are stored\
    \n*** USE THE SAME WORK DIRECTORY FOR ALL DREP OPERATIONS ***")

    Bflags = parent_parser.add_argument_group('SYSTEM PARAMETERS')
    Bflags.add_argument('-p', '--processors', help='threads', default=6, type=int)
    Bflags.add_argument('-d', '--debug', help='make extra debugging output', default=False,
                        action="store_true")
    Bflags.add_argument("-h", "--help", action="help", help="show this help message and exit")

    # Make a parent parser for genome input
    genome_parser = argparse.ArgumentParser(add_help=False)
    Gflags = genome_parser.add_argument_group('GENOME INPUT')
    Gflags.add_argument('-g', '--genomes', nargs='*', help='genomes to filter in .fasta format.\
                                    Not necessary if Bdb or Wdb already exist. Can also input a text file with paths to genomes, which results in fewer OS issues than wildcard expansion')

    # Make a parent parser for filter operation
    filtering_parent = argparse.ArgumentParser(add_help=False)
    fiflags = filtering_parent.add_argument_group('GENOME FILTERING OPTIONS')
    fiflags.add_argument("-l", "--length", help="Minimum genome length", default=50000,
                         type=float)
    fiflags.add_argument("-comp", "--completeness", help="Minimum genome completeness",
                         default=75, type=float)
    fiflags.add_argument("-con", "--contamination", help="Maximum genome contamination",
                         default=25, type=float)

    quality_parent = argparse.ArgumentParser(add_help=False)
    Iflags = quality_parent.add_argument_group('GENOME QUALITY ASSESSMENT OPTIONS')
    Iflags.add_argument("--ignoreGenomeQuality", help="Don't run checkM or do any \
                quality filtering. NOT RECOMMENDED! This is useful for use with bacteriophages\
                or eukaryotes or things where checkM scoring does not work. Will only \
                choose genomes based on length and N50", action='store_true')
    Iflags.add_argument('--genomeInfo', help='location of .csv file containing quality \
                    information on the genomes. Must contain: ["genome"(basename of .fasta file \
                    of that genome), "completeness"(0-100 value for completeness of the genome), \
                    "contamination"(0-100 value of the contamination of the genome)]')
    Iflags.add_argument("--checkM_method", help="Either lineage_wf (more accurate) \
                        or taxonomy_wf (faster)", choices={'taxonomy_wf', 'lineage_wf'}, \
                        default='lineage_wf')
    Iflags.add_argument("--set_recursion", help="Increases the python recursion limit. " \
                                                + "NOT RECOMMENDED unless checkM is crashing due to recursion issues. " \
                                                + "Recommended to set to 2000 if needed, but setting this could crash python", \
                        default='0')
    Iflags.add_argument("--checkm_group_size", help="The number of genomes passed to checkM at a time. Increasing this increases RAM but makes checkM faster", \
                        default=2000, type=int)

    # Make a parent parser for the cluster operation
    cluster_parent = argparse.ArgumentParser(add_help=False)

    Clustflags = cluster_parent.add_argument_group('GENOME COMPARISON OPTIONS')
    Clustflags.add_argument("--S_algorithm", help="R|Algorithm for secondary clustering comaprisons:\n" \
                                                  + "fastANI = Kmer-based approach; very fast\n" \
                                                  + "ANImf   = (DEFAULT) Align whole genomes with nucmer; filter alignment; compare aligned regions\n" \
                                                  + "ANIn    = Align whole genomes with nucmer; compare aligned regions\n" \
                                                  + "gANI    = Identify and align ORFs; compare aligned ORFS\n" \
                                                  + "goANI   = Open source version of gANI; requires nsmimscan\n",
                            default='fastANI', choices={'ANIn', 'gANI', 'ANImf', 'goANI', 'fastANI'})
    Clustflags.add_argument("-ms", "--MASH_sketch", help="MASH sketch size", default=1000)
    Clustflags.add_argument("--SkipMash", help="Skip MASH clustering,\
                            just do secondary clustering on all genomes", action='store_true')
    Clustflags.add_argument("--SkipSecondary", help="Skip secondary clustering, just perform MASH\
                            clustering", action='store_true')
    Clustflags.add_argument("--n_PRESET", help="R|Presets to pass to nucmer\n" \
                                               + "tight   = only align highly conserved regions\n" \
                                               + "normal  = default ANIn parameters", choices=['normal', 'tight'],
                            default='normal')

    Compflags = cluster_parent.add_argument_group('GENOME CLUSTERING OPTIONS')
    Compflags.add_argument("-pa", "--P_ani", help="ANI threshold to form primary (MASH) clusters",
                           default=0.9, type=float)
    Compflags.add_argument("-sa", "--S_ani", help="ANI threshold to form secondary clusters",
                           default=0.95, type=float)
    Compflags.add_argument("-nc", "--cov_thresh", help="Minmum level of overlap between\
        genomes when doing secondary comparisons", default=0.1, type=float)
    Compflags.add_argument("-cm", "--coverage_method", help="R|Method to calculate coverage of an alignment\n" \
                                                            + "(for ANIn/ANImf only; gANI and fastANI can only do larger method)\n"
                                                            + "total   = 2*(aligned length) / (sum of total genome lengths)\n" \
                                                            + "larger  = max((aligned length / genome 1), (aligned_length / genome2))\n",
                           choices=['total', 'larger'], default='larger')
    Compflags.add_argument("--clusterAlg", help="Algorithm used to cluster genomes (passed\
                        to scipy.cluster.hierarchy.linkage", default='average',
                           choices={'single', 'complete', 'average', 'weighted', 'centroid', 'median', 'ward'})

    GRflags = cluster_parent.add_argument_group('GREEDY CLUSTERING OPTIONS\n'
                                                'These decrease RAM use and runtime at the expense of a minor loss in '
                                                'accuracy.\nRecommended when clustering 5000+ genomes')
    GRflags.add_argument("--multiround_primary_clustering",
                         help='Cluster each primary clunk separately and '
                              'merge at the end with single linkage. Decreases '
                              'RAM usage and increases speed, and the cost of a minor '
                              'loss in precision and the inability to plot '
                              'primary_clustering_dendrograms. Especially helpful '
                              'when clustering 5000+ genomes. Will be done with '
                              'single linkage clustering',
                         action='store_true')
    GRflags.add_argument("--primary_chunksize",
                         help="Impacts multiround_primary_clustering. If you have more than this many genomes, "
                              "process them in chunks of this size.",
                         default=5000, type=int)
    GRflags.add_argument("--greedy_secondary_clustering",
                         help="Use a heuristic to avoid pair-wise comparisons when doing secondary clustering. Will "
                              "be done with single linkage clustering. Only works for fastANI S_algorithm option at "
                              "the moment",
                         action='store_true')
    GRflags.add_argument("--run_tertiary_clustering",
                            help="Run an additional round of clustering on the final genome set. This is especially "
                                 "useful when greedy clustering is performed and/or to handle cases where similar genomes "
                                 "end up in different primary clusters. Only works with dereplicate, not compare.",
                            action='store_true')

    # Make a parent parser for scoring
    scoring_parent = argparse.ArgumentParser(add_help=False)
    Sflags = scoring_parent.add_argument_group(
        "SCORING CRITERIA\nBased off of the formula: \nA*Completeness - B*Contamination + C*(Contamination * ("
        "strain_heterogeneity/100)) + D*log(N50) + E*log(size) + F*(centrality - S_ani)\n\nA = completeness_weight; B = "
        "contamination_weight; C = strain_heterogeneity_weight; D = N50_weight; E = size_weight; F = cent_weight")

    Sflags.add_argument("-comW", "--completeness_weight", default=1, type=float,
                        help='completeness weight')
    Sflags.add_argument("-conW", "--contamination_weight", default=5, type=float,
                        help='contamination weight')
    Sflags.add_argument("-strW", "--strain_heterogeneity_weight", default=1, type=float,
                        help='strain heterogeneity weight')
    Sflags.add_argument("-N50W", "--N50_weight", default=0.5, type=float,
                        help='weight of log(genome N50)')
    Sflags.add_argument("-sizeW", "--size_weight", default=0, type=float,
                        help='weight of log(genome size)')
    Sflags.add_argument("-centW", "--centrality_weight", default=1, type=float,
                        help='Weight of (centrality - S_ani)')
    Sflags.add_argument("-extraW", "--extra_weight_table", default=None, type=str,
                        help='Path to a tab-separated file with two-columns, no headers, listing genome and extra score to apply to that genome')

    # Make a parent parser for evaluate
    evaluate_parent = argparse.ArgumentParser(add_help=False)
    Fflags = evaluate_parent.add_argument_group('WARNINGS')
    Fflags.add_argument("--gen_warnings", default=False, help="Generate warnings", action='store_true')
    Fflags.add_argument("--warn_dist", default=0.25, help="How far from the threshold " + \
                                                          " to throw cluster warnings")
    Fflags.add_argument("--warn_sim", default=0.98, help="Similarity threshold for " + \
                                                         " warnings between dereplicated genomes")
    Fflags.add_argument("--warn_aln", default=0.25, help="Minimum aligned fraction for " + \
                                                         " warnings between dereplicated genomes (ANIn)")

    # Make a parent parser for analyze
    analyze_parent = argparse.ArgumentParser(add_help=False)
    Aflags = analyze_parent.add_argument_group('ANALYZE')
    Aflags.add_argument("--skip_plots", default=False, help="Dont make plots", action='store_true')

    dereplicate_parser = subparsers.add_parser("dereplicate", formatter_class=SmartFormatter, \
                                               parents=[parent_parser, genome_parser, filtering_parent, quality_parent,
                                                        cluster_parent, scoring_parent, \
                                                        evaluate_parent, analyze_parent], add_help=False, epilog= \
                                                   "Example: dRep dereplicate output_dir/ -g /path/to/genomes/*.fasta")

    dereplicate_parser = subparsers.add_parser("compare", formatter_class=SmartFormatter, \
                                               parents=[parent_parser, genome_parser, cluster_parent, evaluate_parent], \
                                               add_help=False, epilog= \
                                                   "Example: dRep compare output_dir/ -g /path/to/genomes/*.fasta")

    dep_parser = subparsers.add_parser("check_dependencies", formatter_class=SmartFormatter)

    '''
    ####### PARSE THE ARGUMENTS ######
    '''

    # Handle the situation where the user wants the raw help
    if len(args) == 0 or args[0] == '-h' or args[0] == '--help':
        printHelp()
        sys.exit(0)
    else:
        return parser.parse_args(args)



from setuptools import setup, find_packages
import os

def version():
    setupDir = os.path.dirname(os.path.realpath(__file__))
    versionFile = open(os.path.join(setupDir, 'drep', 'VERSION'))
    return versionFile.read().strip()

setup(name='drep',
      version=version(),
      description='De-replication of microbial genomes assembled from multiple samples',
      url='https://github.com/MrOlm/drep',
      author='Matt Olm',
      author_email='mattolm@berkeley.edu',
      license='MIT',
      package_data={'drep': ['VERSION']},
      packages=find_packages(),
      scripts=['bin/dRep', 'helper_scripts/parse_stb.py', 'helper_scripts/ScaffoldLevel_dRep.py'],
      install_requires=[
          'numpy',
          'pandas',
          'seaborn',
          'matplotlib',
          'biopython',
          'scikit-learn',
          'tqdm',
          'pytest'
      ],
      zip_safe=False)



#!/usr/bin/env python3

###############################################################################
#
# dRep - main program entry point
#
###############################################################################

'''
Controller- takes input from argparse and calls correct modules
'''


__author__ = "Matt Olm"
__license__ = "MIT"
__email__ = "mattolm@gmail.com"
__status__ = "Development"

import argparse
import logging
import os
import sys

import drep
import drep.d_cluster.controller
from drep.WorkDirectory import WorkDirectory
import drep.d_cluster
import drep.d_analyze
import drep.d_filter
import drep.d_choose
import drep.d_adjust
import drep.d_bonus
import drep.d_evaluate
import drep.d_workflows

def version():
    versionFile = open(os.path.join(drep.__path__[0], 'VERSION'))
    return versionFile.read().strip()

VERSION = version()

class Controller():
    def __init__(self):
        self.logger = logging.getLogger()

    def dereplicate_operation(self, **kwargs):
        logging.debug("Starting the dereplicate operation")
        drep.d_workflows.dereplicate_wrapper(kwargs['work_directory'],**kwargs)
        logging.debug("Finished the dereplicate operation!")

    def compare_operation(self, **kwargs):
        logging.debug("Starting the compare operation")
        drep.d_workflows.compare_wrapper(kwargs['work_directory'],**kwargs)
        logging.debug("!!! Finished the compare operation !!!")

    def setup_logger(self,loc):
        ''' set up logger such that DEBUG goes only to file, rest go to file and console '''

        # set up logging everything to file
        logging.basicConfig(level=logging.DEBUG,
                           format='%(asctime)s %(levelname)-8s %(message)s',
                           datefmt='%m-%d %H:%M',
                           filename=loc)

        # set up logging of INFO or higher to sys.stderr
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        formatter = logging.Formatter('%(message)s')
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)

        logging.debug("!"*80)
        logging.debug("***Logger started up at {0}***".format(loc))
        logging.debug("Command to run dRep was: {0}\n".format(' '.join(sys.argv)))
        logging.debug("dRep version {0} was run \n".format(VERSION))
        logging.debug("!"*80 + '\n')

    def parseArguments(self, args):
        ''' Parse user options and call the correct pipeline'''
        if args.operation == 'check_dependencies':
            drep.d_bonus.check_dependencies(print_out=True)
            return

        # Load the workDirectory
        wd_loc = str(os.path.abspath(args.work_directory))
        wd = WorkDirectory(wd_loc)

        # Set up the logger
        self.setup_logger(wd.get_loc('log'))
        logging.debug(str(args))

        # Do some testing
        if args.run_tertiary_clustering:
            if args.operation != "dereplicate":
                raise ValueError("Can only run tertiary clustering with dereplicate")


        # Call the appropriate workflow
        if args.operation == "dereplicate":
            self.dereplicate_operation(**vars(args))
        if args.operation == "compare":
            self.compare_operation(**vars(args))

    def loadDefaultArgs(self):
        pass



#!/usr/bin/env python3
'''
d_choose - a subset of drep

Choose best genome from each cluster
'''

import logging
import glob
import pandas as pd
import os
import sys
import shutil
import numpy as np

import drep
import drep.d_cluster
import drep.d_filter
import drep.WorkDirectory

def d_choose_wrapper(wd, **kwargs):
    '''
    Controller for the dRep choose operation

    Based off of the formula:
    A*Completeness - B*Contamination + C*(Contamination *
    (strain_heterogeneity/100)) + D*log(N50) + E*log(size)

    A = completeness_weight;
    B = contamination_weight;
    C = strain_heterogeneity_weight;
    D = N50_weight;
    E = size_weight

    Args:
        wd (WorkDirectory): The current workDirectory
        **kwargs: Command line arguments

    Keyword args:
        genomeInfo: .csv genomeInfo file
        ignoreGenomeQuality: Don't run checkM or do any quality-based filtering (not recommended)
        checkM_method: Either lineage_wf (more accurate) or taxonomy_wf (faster)

        completeness_weight: see formula
        contamination_weight: see formula
        strain_heterogeneity_weight: see formula
        N50_weight: see formula
        size_weight: see formula

    Returns:
        Makes Sdb (scoreDb) in the workDirectory
    '''
    # Load the WorkDirectory.
    logging.info("Loading work directory")
    workDirectory = drep.WorkDirectory.WorkDirectory(wd)
    logging.debug(str(workDirectory))
    wd = workDirectory

    # Make sure you have enough
    kwargs = _validate_choose_arguments(workDirectory, kwargs)
    Cdb = wd.get_db('Cdb')
    bdb = wd.get_db('Bdb')

    # Get comp/con information
    if kwargs.get('ignoreGenomeQuality', False):
        logging.debug("Skipping all quality-based filtering")
        Gdb = drep.d_filter.calc_genome_info(bdb['location'].tolist())
    else:
        Gdb = drep.d_filter._get_run_genomeInfo(wd, bdb, **kwargs)

    if kwargs.get('centrality_weight', 0) > 0:
        Gdb = add_centrality(wd, Gdb, **kwargs)

    wd.store_db(Gdb, 'genomeInformation', overwrite=True)

    # Call a method with Cdb and Chdb, returning Sdb (scored db) and Wdb (winner db)

    Sdb, Wdb = choose_winners(Cdb, Gdb, **kwargs)

    # Save Sdb and Wdb
    workDirectory.store_db(Sdb,'Sdb',overwrite=True)
    workDirectory.store_db(Wdb,'Wdb',overwrite=True)

    # Make a "winning genomes" folder and populate it
    logging.debug('saving dereplicated genomes')
    g2l = bdb.set_index('genome')['location'].to_dict()
    derep_genomes = [g2l[g] for g in list(Wdb['genome'].unique())]
    wd.store_special('dereplicated_genomes', derep_genomes)

def choose_winners(Cdb, Gdb, **kwargs):
    '''
    Make a scoring database and pick the winner of each cluster

    Args:
        Cdb: clustering database
        Gdb: genome information database

    Keyword Args:
        See wrapper

    Returns:
        List: [Sdb (scoring database), Wdb (winner database)]
    '''
    # If you have extra genome weights, load them
    ew_loc = kwargs.get('extra_weight_table', None)
    if ew_loc is not None:
        Edb = load_extra_weight_table(ew_loc, list(Cdb['genome'].unique()), **kwargs)
    else:
        Edb = None

    # Generate Sdb
    genomes = list(Cdb['genome'].unique())
    Sdb = score_genomes(genomes, Gdb, Edb=Edb, **kwargs)
    logging.debug("Sdb finished")

    # Generate Wdb
    Wdb = pick_winners(Sdb,Cdb)
    logging.debug("Wdb finished")

    return Sdb, Wdb

def load_extra_weight_table(loc, genomes, **kwargs):
    """

    Args:
        loc: location of extra weight table
        genomes: list of genomes you have in Cdb
        **kwargs: nothing realld

    Returns:
        dataframe with columns "genome" and "extra_weight"

    """
    if not os.path.join(loc):
        logging.error(f"COULD NOT FIND FILE {loc}; WILL NOT PROCESS EXTRA WEIGHTS")
        return None
    else:
        try:
            db = pd.read_csv(loc, sep='\t', names=['genome', 'extra_weight'])
            db['extra_weight'] = db['extra_weight'].astype(float)
        except:
            f = ''
            with open(loc, 'r') as o:
                x = 0
                for line in o.readlines():
                    f += line + '\n'
                    x += 1

                    if x > 10:
                        break

            logging.error(f"COULD NOT LOAD FILE {loc}; WILL NOT PROCESS EXTRA WEIGHTS. FILE LOOKS LIKE:\n{f}")
            return None

        wg = set(db['genome'].tolist())
        gg = set(genomes)

        missing = "\n".join(list(wg-gg)[:10])

        logging.info(f'Loaded {len(db)} extra weights. {len(gg.intersection(wg))} of {len(gg)} genomes have an extra weight. {len(wg - gg)} genomes have a weight but ARE NOT KNOWN BY DREP; here are some examples:\n{missing}')

        return db[db['genome'].isin(gg)]

def pick_winners(Sdb, Cdb):
    '''
    Based on clustering and scores, pick the best genome from every cluster

    Args:
        Sdb: score of every genome
        Cdb: clustering

    Returns:
        DataFrame: Wdb (winner database)
    '''
    Table = {'genome':[],'cluster':[],'score':[]}
    for cluster in Cdb['secondary_cluster'].unique():
        d = Sdb[Sdb['genome'].isin(Cdb['genome'][Cdb['secondary_cluster'] == cluster].tolist())]
        score = d['score'].max()
        genome = d['genome'][d['score'] == score].tolist()[0]
        Table['genome'].append(genome)
        Table['score'].append(score)
        Table['cluster'].append(cluster)

    Wdb = pd.DataFrame(Table)
    return Wdb

def score_genomes(genomes, Gdb, Edb=None, **kwargs):
    '''
    Calculate the scores for a list of genomes

    Args:
        genomes: list of genomes
        Gdb: genome information database

    Keyword Args:
        See wrapper

    Returns:
        DataFrame: Sdb (scoring database)
    '''

    if Edb is not None:
        g2e = Edb.set_index('genome')['extra_weight'].to_dict()
    else:
        g2e = {}

    Table = {'genome':[],'score':[]}
    for genome in genomes:
        if genome in g2e:
            extra = g2e[genome]
        else:
            extra = 0
        row = Gdb[Gdb['genome'] == genome]

        score = score_row(row, extra=extra, **kwargs)
        Table['genome'].append(genome)
        Table['score'].append(score)

    Sdb = pd.DataFrame(Table)
    return Sdb

def score_row(row, extra=0, **kwargs):
    '''
    Perform the scoring of a row based on kwargs

    Args:
        row: row of genome information

    Keyword Args:
        ignoreGenomeQuality: Don't run checkM or do any quality-based filtering (not recommended)

        completeness_weight: see formula
        contamination_weight: see formula
        strain_heterogeneity_weight: see formula
        N50_weight: see formula
        size_weight: see formula
        extra = extra weight to apply

    Returns:
        float: score
    '''
    comW =  kwargs.get('completeness_weight',1)
    conW =  kwargs.get('contamination_weight',1)
    n50W =  kwargs.get('N50_weight',1)
    sizeW = kwargs.get('size_weight',1)
    strW =  kwargs.get('strain_heterogeneity_weight',1)
    centW = kwargs.get('centrality_weight', 0)

    # For centrality calculations
    S_ani = kwargs.get('S_ani', 0.99)
    if centW > 0:
        cent = row['centrality'].tolist()[0]
    else:
        cent = 0

    n50 = float(row['N50'].tolist()[0])
    size = float(row['length'].tolist()[0])

    if kwargs.get('ignoreGenomeQuality', False):
        score = (np.log10(n50) * n50W) + (np.log10(size) * sizeW) + ((cent - S_ani) * centW) + float(extra)
        return score

    com = float(row['completeness'].tolist()[0])
    con = float(row['contamination'].tolist()[0])

    if 'strain_heterogeneity' in row:
        strh = float(row['strain_heterogeneity'].tolist()[0])
    else:
        strh = 0

    score = (com * comW) - (con * conW) + (strW * (con * (strh/100))) \
        + (np.log10(n50) * n50W) + (np.log10(size) * sizeW) + ((cent - S_ani) * centW) + float(extra)
    return score

def _validate_choose_arguments(wd, kwargs):
    '''
    Validate choose arguments

    Make sure you have a Cdb

    Args:
        wd: WorkDirectory
        kwargs: keyword arguments
    '''
    if not wd.hasDb('Cdb'):
        logging.error("Can't find Cdb- quitting")
        logging.error("Cdb is not found in the work directory- you must run cluster before you ",
                + "can choose")
        sys.exit()

    # Validate centrality arguments
    if (kwargs.get('SkipSecondary', True)) & (kwargs.get('centrality_weight', 0) > 0):
        logging.error(
            "You skipped secondary clustering but have centrality weight above 0. You cant do that. I will now set the centrality weight to 0 to avoid a crash")
        kwargs['centrality_weight'] = 0

    return kwargs

def add_centrality(wd, Gdb, **kwargs):
    """
    Add a columns named "centrality" to genome info
    """
    Ndb = wd.get_db('Ndb')
    Cdb = wd.get_db('Cdb')

    if Cdb['cluster_method'].iloc[0] == 'greedy':
        Ndb = calc_centrality_from_scratch(wd.get_db('Bdb'), Cdb, os.path.join(wd.get_dir('MASH'), 'centrality_calculations/'))

    g2c = Cdb.set_index('genome')['secondary_cluster'].to_dict()
    c2s = Cdb['secondary_cluster'].value_counts().to_dict()

    Ndb['cluster_1'] = Ndb['reference'].map(g2c)
    Ndb['cluster_2'] = Ndb['querry'].map(g2c)
    Ndb = Ndb[Ndb['cluster_1'] == Ndb['cluster_2']]
    Ndb = Ndb[Ndb['reference'] != Ndb['querry']]

    genome2centrality = {}
    for cluster, ndb in Ndb.groupby('cluster_1'):
        #print(f"Cluster {cluster} has {c2s[cluster]} members and {len(ndb)} comps")

        mlen = c2s[cluster]
        assert len(ndb) == (mlen * mlen) - mlen
        for genome, db in ndb.groupby('reference'):
            genome2centrality[genome] = db['ani'].mean()

    Gdb['centrality'] = Gdb['genome'].map(genome2centrality).fillna(0)
    return Gdb

def calc_centrality_from_scratch(Bdb, Cdb, data_folder):
    """
    Calculate centrality from scratch using Mash
    """
    logging.info("Calculating centrality using Mash")

    # 1) Run calculations
    dbs = []
    Xdb = pd.merge(Cdb, Bdb, on='genome', how='left')
    for cluster, bdb in Xdb.groupby('secondary_cluster'):
        if len(bdb) <= 1:
            continue

        df = os.path.join(data_folder, cluster + '/')
        mdb, cdb, cluster_ret = drep.d_cluster.compare_utils.all_vs_all_MASH(bdb, df, MASH_sketch=10000)
        mdb['ani'] = 1 - mdb['dist']
        mdb['cluster'] = cluster
        mdb = mdb.rename(columns={'genome1':'reference', 'genome2':'querry'})
        dbs.append(mdb[['reference', 'querry', 'ani', 'cluster']])

    Mdb = pd.concat(dbs).reset_index(drop=True)
    return Mdb



import glob
import logging
import os
import shutil
import sys

import numpy as np
import pandas as pd

import drep
import drep.d_cluster.utils
# from drep.d_cluster.utils import _get_genome_name_from_fasta, process_deltafiles, _randomString, gen_gANI_cmd, \
#     process_gani_files, gen_goANI_cmd, process_goani_files

def run_pairwise_ANImf(genome_list, ANIn_folder, **kwargs):
    '''
    Given a list of genomes and an output folder, compare all genomes using ANImf

    Args:
        genome_list: list of locations of genome files
        ANIn_folder: folder to store the output of comparison

    Keyword arguments:
        processors: threads to use
        debug: if true save extra output
        wd: needed if debug is True
    '''
    p = kwargs.get('processors',6)
    genomes = genome_list

    # Make folder if doesnt exist
    if not os.path.exists(ANIn_folder):
        os.makedirs(ANIn_folder)

    # Gen commands
    cmds = []
    files = []
    for g1 in genomes:
        # Make it so each reference is it's own folder, to spread out .delta files
        cur_folder = os.path.join(ANIn_folder, drep.d_cluster.utils._get_genome_name_from_fasta(g1))
        if not os.path.exists(cur_folder):
            os.makedirs(cur_folder)

        for g2 in genomes:
            file_name = "{0}/{1}_vs_{2}".format(cur_folder,
                        drep.d_cluster.utils._get_genome_name_from_fasta(g1),
                        drep.d_cluster.utils._get_genome_name_from_fasta(g2))
            files.append(file_name)

            # If the file doesn't already exist, add it to what needs to be run
            if not os.path.isfile(file_name + '.delta.filtered'):
                cmds.append(gen_animf_cmd(file_name,g1,g2))

    # Run commands
    if len(cmds) > 0:
        for c in cmds:
            logging.debug(c)

        if ('wd' in kwargs) and (kwargs.get('debug', False)):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
        drep.thread_cmds(cmds, shell=True, logdir=logdir, t=int(p))

    # Make dictionary of genome lengths
    org_lengths = {}
    for genome in genomes:
        org_lengths[drep.d_cluster.utils._get_genome_name_from_fasta(genome)] = drep.d_filter.calc_fasta_length(genome)

    # Parse output
    deltafiles = ["{0}.delta.filtered".format(f) for f in files]
    df = drep.d_cluster.utils.process_deltafiles(deltafiles, org_lengths, **kwargs)

    return df


def run_pairwise_fastANI(genome_list, outdir, **kwargs):
    p = kwargs.get('processors',6)
    code = drep.d_cluster.utils._randomString(stringLength=10)

    # Make folders
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    tmp_dir = os.path.join(outdir, 'tmp/')
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    # Make genome list
    glist = os.path.join(tmp_dir, 'genomeList')
    glist = _make_glist(genome_list, glist)

    # Gen command
    exe_loc = drep.get_exe('fastANI')
    out_base = os.path.join(outdir, 'fastANI_out_{0}'.format(code))
    cmd = [exe_loc, '--ql', glist, '--rl', glist, '-o', out_base, '--matrix', '-t', str(p), "--minFraction", str(0)]
    logging.debug(' '.join(cmd) + ' ' + code)

    # Run command
    if ('wd' in kwargs) and (kwargs.get('debug', False)):
        logdir = kwargs.get('wd').get_dir('cmd_logs')
    else:
        logdir = False
    drep.thread_cmds([cmd], shell=False, logdir=logdir, t=1)

    # Load results
    fdb = load_fastani(out_base)

    # fix missing ones
    try:
        fdb = _fix_fastani(fdb)
        return fdb

    # handle broken self
    except:
        logging.error("CRITICAL ERROR WITH SECONDARY CLUSTERING CODE {0}; SKIPPING".format(code))
        return pd.DataFrame()

def fastani_one_vs_many(one, many, genome_rep_file, outdir, **kwargs):
    p = kwargs.get('processors', 6)
    code = drep.d_cluster.utils._randomString(stringLength=10)
    tmp_dir = kwargs.get('tmp_dir')
    logdir = kwargs.get('logdir')
    exe_loc = kwargs.get('current_exe')
    redo = kwargs.get('redo', False)

    # Gen command
    out_base = os.path.join(outdir, 'fastANI_out_{0}'.format(code))
    cmd = [exe_loc, '-q', one, '--rl', genome_rep_file, '-o', out_base, '--matrix', '-t', str(p), "--minFraction", str(0)]
    logging.debug(' '.join(cmd) + ' ' + code)

    # Run command
    drep.thread_cmds([cmd], shell=False, logdir=logdir, t=1)

    # Load results
    fdb = load_fastani(out_base)

    return fdb

def load_fastani(file):
    fdb = pd.read_csv(file, names=['reference', 'querry', 'ANI', 'j1', 'j2'], delim_whitespace=True)
    for c in ['reference', 'querry']:
        fdb[c] = [drep.d_cluster.utils._get_genome_name_from_fasta(x) for x in fdb[c]]
    fdb = fdb.rename(columns={'ANI':'ani'})
    fdb['alignment_coverage'] = [(j1/j2) for j1, j2 in zip(fdb['j1'], fdb['j2'])]
    fdb = fdb[['reference', 'querry', 'ani', 'alignment_coverage']]
    fdb['ani'] = [x/100 for x in fdb['ani']]

    return fdb


def _fix_fastani(odb):
    # Add back missing genomes
    fdb = odb.pivot('reference', 'querry', 'ani')
    fdb.reset_index(level=0, inplace=True)
    fdb.fillna(0, inplace=True)
    fdb = fdb.melt(id_vars=['reference']).rename(
            columns={'value':'ani'})

    # Add back alignment coverage
    fdb = pd.merge(fdb, odb[['reference', 'querry', 'alignment_coverage']], on=['reference', 'querry'], how='outer')
    fdb['alignment_coverage'] = fdb['alignment_coverage'].fillna(0)

    assert len(fdb['reference'].unique()) == len(fdb['querry'].unique())
    assert len(fdb) == (len(fdb['reference'].unique()) * len(fdb['querry'].unique()))

    return fdb


def _make_glist(genomes, floc):
    o = open(floc, 'w')
    for g in genomes:
        loc = os.path.abspath(g)
        assert os.path.isfile(g)
        o.write(loc + '\n')
    o.close()
    return floc


def run_pairwise_gANI(bdb, gANI_folder, prod_folder, **kwargs):
    '''
    Run pairwise gANI on a list of Genomes

    Args:
        bdb: DataFrame with ['genome', 'location']
        gANI_folder: folder to store gANI output
        prod_folder: folder containing prodigal output from genomes (will run if needed)

    Keyword arguments:
        debug: log all of the commands
        wd: if you want to log commands, you also need the wd
        processors: threads to use

    Returns:
        DataFrame: Ndb for gANI
    '''
    p = kwargs.get('processors',6)
    gANI_exe = drep.get_exe('ANIcalculator')
    genomes = bdb['location'].tolist()

    # Make folders
    if not os.path.exists(gANI_folder):
        os.makedirs(gANI_folder)
    if not os.path.exists(prod_folder):
        os.makedirs(prod_folder)

    # Remove crap folders- they shouldn't exist and if they do it messes things up
    crap_folders = glob.glob(gANI_folder + '*.gANITEMP')
    for crap_folder in crap_folders:
        if os.path.exists(crap_folder):
            shutil.rmtree(crap_folder)

    # Run prodigal
    logging.debug("Running prodigal...")
    drep.d_filter.run_prodigal(bdb['location'].tolist(), prod_folder, **kwargs)

    # Gen gANI commands
    logging.debug("Running gANI...")
    cmds = []
    files = []
    for i, g1 in enumerate(genomes):
        # Make it so each reference is it's own folder, to spread out .delta files
        cur_folder = os.path.join(gANI_folder, drep.d_cluster.utils._get_genome_name_from_fasta(g1))
        if not os.path.exists(cur_folder):
            os.makedirs(cur_folder)

        for j, g2 in enumerate(genomes):
            if i > j:
                name1= drep.d_cluster.utils._get_genome_name_from_fasta(g1)
                name2= drep.d_cluster.utils._get_genome_name_from_fasta(g2)
                file_name = "{0}/{1}_vs_{2}.gANI".format(cur_folder,
                            name1, name2)
                files.append(file_name)

                # If the file doesn't already exist, add it to what needs to be run
                if not os.path.isfile(file_name):
                    fna1 = "{0}.fna".format(os.path.join(prod_folder,name1))
                    fna2 = "{0}.fna".format(os.path.join(prod_folder,name2))
                    cmds.append(drep.d_cluster.utils.gen_gANI_cmd(file_name,fna1,fna2,gANI_exe))

    # Run commands
    if len(cmds) > 0:
        logging.debug('Running gANI commands: {0}'.format('\n'.join([' '.join(x) for x in cmds])))
        if ('wd' in kwargs) and (kwargs.get('debug', False) == True):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
            #logdir = "/home/mattolm/Programs/drep/tests/test_backend/logs/"
        drep.thread_cmds(cmds, logdir=logdir, t=int(p))

    else:
        logging.debug("gANI already run- will not re-run")

    # Pdrep.d_cluster.utils.arse output
    df = drep.d_cluster.utils.process_gani_files(files)

    # Add self-comparisons if there is only one genome
    if len(genomes) == 1:
        Table = {'querry':[],'reference':[],'ani':[],'alignment_coverage':[]}
        for g in genomes:
            Table['reference'].append(drep.d_cluster.utils._get_genome_name_from_fasta(g))
            Table['querry'].append(drep.d_cluster.utils._get_genome_name_from_fasta(g))
            Table['ani'].append(1)
            Table['alignment_coverage'].append(1)
        d = pd.DataFrame(Table)
        df = pd.concat([df,d],ignore_index=True)

    return df


def run_pairwise_goANI(bdb, goANI_folder, prod_folder, **kwargs):
    '''
    Run pairwise goANI on a list of Genomes

    Args:
        bdb: DataFrame with ['genome', 'location']
        goANI_folder: folder to store gANI output
        prod_folder: folder containing prodigal output from genomes (will run if needed)

    Keyword arguments:
        debug: log all of the commands
        wd: if you want to log commands, you also need the wd
        processors: threads to use

    Returns:
        DataFrame: Ndb for gANI
    '''
    p = kwargs.get('processors',6)
    nsimscan_exe = drep.get_exe('nsimscan')
    genomes = bdb['location'].tolist()

    # Make folders
    if not os.path.exists(goANI_folder):
        os.makedirs(goANI_folder)
    if not os.path.exists(prod_folder):
        os.makedirs(prod_folder)

    # Run prodigal
    logging.debug("Running prodigal...")
    drep.d_filter.run_prodigal(bdb['location'].tolist(), prod_folder, **kwargs)

    # Gen gANI commands
    logging.debug("Running goANI...")
    cmds = []
    files = []
    for i, g1 in enumerate(genomes):
        # Make it so each reference is it's own folder, to spread out .delta files
        cur_folder = os.path.join(goANI_folder, drep.d_cluster.utils._get_genome_name_from_fasta(g1))
        if not os.path.exists(cur_folder):
            os.makedirs(cur_folder)

        for j, g2 in enumerate(genomes):
            if i != j:
                name1= drep.d_cluster.utils._get_genome_name_from_fasta(g1)
                name2= drep.d_cluster.utils._get_genome_name_from_fasta(g2)
                file_name = "{0}/{1}_vs_{2}.sim".format(cur_folder,
                            name1, name2)
                files.append(file_name)

                # If the file doesn't already exist, add it to what needs to be run
                if not os.path.isfile(file_name):
                    fna1 = "{0}.fna".format(os.path.join(prod_folder,name1))
                    fna2 = "{0}.fna".format(os.path.join(prod_folder,name2))
                    cmds.append(drep.d_cluster.utils.gen_goANI_cmd(file_name,fna1,fna2,nsimscan_exe))

    # Run commands
    if len(cmds) > 0:
        logging.debug('Running goANI commands: {0}'.format('\n'.join([' '.join(x) for x in cmds])))
        if ('wd' in kwargs) and (kwargs.get('debug', False) == True):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
            #logdir = "/home/mattolm/Programs/drep/tests/test_backend/logs/"
        drep.thread_cmds(cmds, logdir=logdir, t=int(p))

    else:
        logging.debug("goANI already run- will not re-run")

    # Parse output
    df = drep.d_cluster.utils.process_goani_files(files)

    # Add self-comparisons if there is only one genome
    if len(genomes) == 1:
        Table = {'querry':[],'reference':[],'ani':[],'alignment_coverage':[]}
        for g in genomes:
            Table['reference'].append(drep.d_cluster.utils._get_genome_name_from_fasta(g))
            Table['querry'].append(drep.d_cluster.utils._get_genome_name_from_fasta(g))
            Table['ani'].append(1)
            Table['alignment_coverage'].append(1)
        d = pd.DataFrame(Table)
        df = pd.concat([df,d],ignore_index=True)

    return df


def parse_gani_file(file):
    '''
    Parse gANI file, return dictionary of results

    Args:
        file: location of gANI file

    Returns:
        dict: results in the gANI file
    '''
    try:
        x = pd.read_table(file)
    except:
        logging.error('gANI file {0} does not exist. The most likely reason is that '.format(file)\
            + 'one of the genomes has a .fasta header than gANI doesnt like.'\
            + ' Known issues include having a header over 160 characaters long, or any '\
            + 'kind of special character besides "_" (including ".", ":", and "-").'\
            + ' To fix this error either use ANIn (which is not as picky with fasta headers)'\
            + ', or fix the .fasta headers to conform to those rules.')
        sys.exit()
    x = x.rename(columns={'GENOME1':'reference','GENOME2':'querry','AF(1->2)':'rq_coverage',
                        'AF(2->1)':'qr_coverage','ANI(1->2)':'rq_ani','ANI(2->1)':'qr_ani'})
    dict = x.to_dict(orient='list')
    dict = {k:v[0] for k,v in dict.items()}
    dict['reference'] = dict['reference'][:-4]
    dict['querry'] = dict['querry'][:-4]
    return dict


def parse_nsim_file(file):
    '''
    Parse nsim file, return dictionary of results and gene datatable

    Args:
        file: location of nsimscan file

    Returns:
        dict: results in the gANI file
        db: gene-based alignment results
    '''
    # Load
    db = pd.read_csv(file, sep='\t')

    if len(db) == 0:
        logging.warning("File {0} is empty, indicating a nsimscan failure! Run with --debug and check the log folder for details".format(file))

        return _summarize_nsimsan(db)

    db = db.rename(columns={'#qry_id':'qry_id', '#Q_id':'qry_id', 'S_id':'sbj_id', 'trg_len':'sbj_len'})

    # Filter
    try:
        db = _filter_nsimscan(db)
    except:
        print("ERROR! FILE {0}".format(file))
        print(db)
        return _summarize_nsimsan(pd.DataFrame())

    # Make summary results
    x = _summarize_nsimsan(db)

    return x


def _filter_nsimscan(db1, minAF=0.7, minANI=70):
    '''
    Filter a single nsim scan file
    '''
    db1 = db1.sort_values(['al_len','qry_id', 'sbj_id'], ascending=False)

    # Filter like gANI
    db1['af'] = [a/min(o,t) for a,o,t in zip(db1['al_len'],
                                             db1['qry_len'],
                                             db1['sbj_len'])]
    db1 = db1[(db1['af'] >= minAF) & (db1['p_inden'] >= minANI)]

    return db1


def _summarize_nsimsan(db):
    '''
    Take all those aligned genes and return a summary ANI
    '''
    table = {}
    if len(db) > 0:
        table['ani'] = sum([p * l for p, l in zip(db['p_inden'], db['al_len'])]) \
                        / db['al_len'].sum()
        table['af'] = db['al_len'].sum() / db['qry_len'].sum()
    else:
        table['ani'] = 0
        table['af'] = 0

    return table


def gen_nucmer_cmd(prefix,ref,querry,c='65',noextend=False,maxgap='90',method='mum'):
    '''
    Generate command to run with nucmer

    Args:
        prefix: desired output name (will have .delta appended)
        ref: location of reference genome
        querry: location of querry genomes

    Keyword args:
        c: c value
        noextend: either True or False
        maxgap: maxgap
        method: detault is 'mum'

    Returns:
        list: command to run number
    '''
    cmd = ['nucmer','--' + method,'-p',prefix,'-c',str(c),'-g',str(maxgap)]
    if noextend: cmd.append('--noextend')
    cmd += [ref,querry]
    return cmd


def gen_animf_cmd(prefix,ref,querry, **kwargs):
    '''
    return animf command. It will be a single string, with the format:
    "nucmer cmd ; filter cmd"

    Args:
        prefix: desired file output name
        ref: location of reference genome
        querry: location of querry genome

    Keyword args:
        all: passed on to gen_nucmer_cmd

    Returns:
        string: format is "nucmer cmd ; filter cmd"
    '''

    nucmer_cmd = gen_nucmer_cmd(prefix,ref,querry, **kwargs)

    delta = prefix + '.delta'
    out = delta + '.filtered'

    filter_cmd = gen_filter_cmd(prefix + '.delta', out)
    return ' '.join(nucmer_cmd) + '; ' + ' '.join(filter_cmd)


def gen_filter_cmd(delta, out):
    '''
    return delta-filter command

    Args:
        delta: desired .delta file to filter
        out: desired output file

    Returns:
        list: cmd to run
    '''
    cmd = ["delta-filter", '-r', '-q', delta, '>', out]
    return cmd


def _gen_nomash_cdb(Bdb):
    '''
    From Bdb, just add a column of 'primary_cluster' = 0

    Args:
        Bdb: dataframe with [genome, location]

    Returns:
        DataFrame: Cdb
    '''
    Cdb = Bdb.copy()
    Cdb['primary_cluster'] = 0
    return Cdb


def add_avani(db):
    '''
    add a column titled 'av_ani' to the passed in dataframe

    dataframe must have rows reference, querey, and ani

    Args:
        db: dataframe
    '''

    logging.debug('making dictionary for average_ani')
    combo2value = {}
    for i, row in db.iterrows():
        combo2value["{0}-vs-{1}".format(row['querry'], row['reference'])] \
            = row['ani']

    logging.debug('list comprehension for average_ani')
    db['av_ani'] = [np.mean([combo2value["{0}-vs-{1}".format(q, r)],
                        combo2value["{0}-vs-{1}".format(r, q)]]) if r != q else 1\
                        for q, r in zip(db['querry'].tolist(),
                        db['reference'].tolist())]

    logging.debug('averageing done')


def _nucmer_preset(preset):
    '''
    Return the values from a nucmer 'preset'

    Args:
        preset: either "tight" or "normal"

    Returns:
        list: [c, n_maxgap, n_noextend, n_method]
    '''
    assert preset in ['tight','normal']

    if preset == 'tight':
        return 65, 1, True, 'mum'

    elif preset == 'normal':
        return 65, 90, False, 'mum'


#!/usr/bin/env python3

import logging
import os
import pandas as pd
import sys

import drep
import drep.WorkDirectory
import drep.d_cluster.compare_utils
import drep.d_cluster.controller
import drep.d_filter
import drep.d_cluster
import drep.d_analyze

def d_evaluate_wrapper(wd,**kwargs):

    logging.debug("Loading work directory")
    wd = drep.WorkDirectory.WorkDirectory(wd)
    logging.debug(str(wd))

    # Run tertiary clustering if that's what you're here to do
    if kwargs.get('run_tertiary_clustering', False):
        logging.info("Running tertiary clustering on genome representatives")
        run_tertiary_clustering(wd, **kwargs)

    # Determine what to evaluate
    evs = kwargs.get('evaluate')
    options = ['1','2','3']
    to_eval = drep.d_analyze._parse_plot_options(options, evs)
    logging.debug("evaluating {0}".format(to_eval))

    # 1) Evaluate de-replicated genome similarity
    if '1' in to_eval:
        logging.info('will compare winners')
        Wmdb, Wndb = compare_winners(wd,**kwargs)

        # Save databases
        wd.store_db(Wmdb,'Wmdb',overwrite=True)
        wd.store_db(Wndb,'Wndb',overwrite=True)

    # 2) Throw warnings for clusters that were almost different
    if '2' in to_eval:
        logging.info('will provide warnings about clusters')
        warnings = evaluate_warnings(wd, **kwargs)

        # Save it
        warn_log = os.path.join(wd.location, 'log/warnings.txt')
        with open(warn_log, 'w') as file:
            file.write('\n'.join(warnings))
            file.write('\n')
        logging.info("{0} warnings generated: saved to {1}".format(len(warnings),warn_log))

    # 3) Generate a database of information on winning genomes
    if '3' in to_eval:
        logging.info('will produce Widb (winner information db)')
        Widb = evaluate_winners(wd, **kwargs)

        # Save it
        wd.store_db(Widb,'Widb',overwrite=True)
        loc = wd.location + 'data_tables/Widb.csv'
        logging.info("Winner database saved to {0}".format(loc))

    return

def compare_winners(wd, **kwargs):
    Bdb = wd.get_db('Bdb')
    Wdb = wd.get_db('Wdb')
    Bdb = Bdb[Bdb['genome'].isin(Wdb['genome'].tolist())]
    data_folder = os.path.join(wd.location, 'data/')
    comp_method = kwargs.get('comp_method','ANIn')

    # Generate MASH db
    Wmdb = drep.d_cluster.compare_utils.all_vs_all_MASH(Bdb, data_folder)

    # Generate ANIn db
    Wndb = drep.d_cluster.compare_utils.compare_genomes(Bdb, comp_method, wd, **kwargs)

    return Wmdb, Wndb

def evaluate_warnings(wd, **kwargs):
    Cdb = wd.get_db('Cdb')
    Ndb = wd.get_db('Ndb')
    warn_dist = float(kwargs.get('warn_dist',.25))

    if 'Blank' in Ndb:
        logging.error("Ndb is blank (was run with --SkipSecondary) - "\
                + "will skip cluster evaluation")

    if wd.hasDb('Wmdb'):
        Wmdb = wd.get_db('Wmdb')
        Wndb = wd.get_db('Wndb')
        warn_sim = float(kwargs.get('warn_sim',.98))
        warn_aln = float(kwargs.get('warn_aln',.25))

    warnings = []

    # Check for cases where there is a close clustering
    # That is, a case where a cluster was almost split or not-split
    for cluster in sorted(Cdb['primary_cluster'].unique()):
        d = Cdb[Cdb['primary_cluster'] == cluster]

        if 'Blank' in Ndb:
            continue

        # Skip if it's a singleton
        if len(d['genome'].unique()) == 1:
            continue

        # Load clustering information
        linkI = wd.get_cluster("secondary_linkage_cluster_{0}".format(cluster))
        db = linkI['db']
        linkage = linkI['linkage']
        args = linkI['arguments']
        threshold = args['linkage_cutoff']
        threshold = (1-threshold) * 100
        names = list(db.columns)

        if linkage is None:
            continue

        # Find cases where the cluster is close to the treshold used
        for node in linkage:
            t = node[2]
            x = (1-t)*100
            if abs(x - threshold) < warn_dist:

                # This means that it wasn't split, but was almost
                if x < threshold:
                    #c1 = Cdb['secondary_cluster'][Cdb['genome'] == names[node[0]]].tolist()[0]
                    #c2 = Cdb['secondary_cluster'][Cdb['genome'] == names[node[1]]].tolist()[0]

                    warning = "CLUSTERING WARNING: Primary cluster {0} was almost not split".\
                            format(cluster)
                    warnings.append(warning)

                # This means that a cluster was split, but almost not
                if x > threshold:
                    #print(node)
                    #c1 = Cdb['secondary_cluster'][Cdb['genome'] == names[int(node[0])]].tolist()[0]
                    #c2 = Cdb['secondary_cluster'][Cdb['genome'] == names[int(node[1])]].tolist()[0]
                    #assert c1 == c2

                    warning = "CLUSTERING WARNING: Primary cluster {0} was almost split".\
                            format(cluster)
                    warnings.append(warning)

        # Find cases where you have a high self comparison
        self_thresh = (1-drep.d_analyze.get_highest_self(Ndb, names))*100
        if self_thresh <= threshold + warn_dist:
            warning = "CLUSTERING WARNING: Primary cluster {0} has a high self-comparison value".\
                            format(cluster)
            warnings.append(warning)

    # Check for cases where winners are very similar
    # Either based on MASH or ANIn
    if wd.hasDb('Wmdb'):
        # See if any MASH comparisons are too similar
        Wmdb = Wmdb[(Wmdb['genome1'] != Wmdb['genome2']) & (Wmdb['similarity'] > warn_sim)\
                   & (Wmdb['genome1'] > Wmdb['genome2'])]
        for i,row in Wmdb.iterrows():
            c1 = Cdb['secondary_cluster'][Cdb['genome'] == row['genome1']].tolist()[0]
            c2 = Cdb['secondary_cluster'][Cdb['genome'] == row['genome2']].tolist()[0]

            warning = "WINNER WARNING: Genomes {0} ({3}) and {1} ({4}) have a high MASH score ({2:.2f}%)".format(\
                        row['genome1'], row['genome2'], row['similarity']*100, c1, c2)
            warnings.append(warning)

        # See if any secondary comparisons are too similar
        Wndb = Wndb[(Wndb['reference'] != Wndb['querry']) & (Wndb['ani'] > warn_sim)\
                   & (Wndb['reference'] > Wndb['querry']) & (Wndb['alignment_coverage'] > warn_aln)]
        for i,row in Wndb.iterrows():
            c1 = Cdb['secondary_cluster'][Cdb['genome'] == row['reference']].tolist()[0]
            c2 = Cdb['secondary_cluster'][Cdb['genome'] == row['querry']].tolist()[0]

            warning = "WINNER WARNING: Genomes {0} ({3}) and {1} ({4}) have a high ANIn score ({2:.2f}% ANI ".format(\
                        row['reference'], row['querry'], row['ani']*100, c1, c2) + "; {0:.2f}% aligned)".format(row['alignment_coverage'])
            warnings.append(warning)

    return warnings

def comp_str(val):
    if val == 100:
        return('perfect')
    if val > 90:
        return("near")
    if val > 70:
        return("substantial")
    if val > 50:
        return("moderate")
    if val <= 50:
        return("partial")

def con_str(val):
    if val == 0:
        return ("none")
    if val <= 5:
        return("low")
    if val < 10:
        return ("medium")
    if val < 15:
        return("high")
    if val >= 15:
        return("very high")


def evaluate_winners(wd, **kwrags):
    Wdb = wd.get_db('Wdb')
    Cdb = wd.get_db('Cdb')
    Ndb = wd.get_db('Ndb')

    # For every winning genome, give some key stats based on the information available
    Table = {'genome':[],'score':[],'completeness':[],'contamination':[],'strain_heterogeneity':[],\
            'size':[],'N50':[],'cluster':[], 'taxonomy':[], 'tax_confidence':[],\
            'cluster_members':[],'closest_cluster_member':[],'furthest_cluster_member':[],\
            'completeness_metric':[], 'contamination_metric':[]}

    for i, row in Wdb.iterrows():
        # Add info in Wdb
        Table['genome'].append(row['genome'])
        Table['score'].append(row['score'])
        Table['cluster'].append(row['cluster'])

        # Add clustering info
        if 'Blank' not in Ndb:
            d = Cdb[Cdb['secondary_cluster'] == row['cluster']]
            ndb = Ndb[(Ndb['reference'] == row['genome']) & (Ndb['querry'].isin(d['genome'].tolist()))\
                     & (Ndb['reference'] != Ndb['querry'])]
            members = len(d['genome'].unique())

            if members > 1:
                Table['closest_cluster_member'].append("{0:.2f}".format(ndb['ani'].max()*100))
                Table['furthest_cluster_member'].append("{0:.2f}".format(ndb['ani'].min()*100))
                Table['cluster_members'].append(len(d['genome'].unique()))
            else:
                Table['closest_cluster_member'].append("NA")
                Table['furthest_cluster_member'].append("NA")
                Table['cluster_members'].append(len(d['genome'].unique()))
        else:
            d = Cdb[Cdb['secondary_cluster'] == row['cluster']]
            Table['closest_cluster_member'].append("NA")
            Table['furthest_cluster_member'].append("NA")
            Table['cluster_members'].append(len(d['genome'].unique()))

        # Add checkM info
        if wd.hasDb('Chdb'):
            Chdb = wd.get_db('Chdb')
            d = Chdb[Chdb['Bin Id'] == row['genome']]
            Table['completeness'].append(d['Completeness'].tolist()[0])
            Table['contamination'].append(d['Contamination'].tolist()[0])
            Table['strain_heterogeneity'].append(d['Strain heterogeneity'].tolist()[0])
            Table['size'].append(d['Genome size (bp)'].tolist()[0])
            Table['N50'].append(d['N50 (scaffolds)'].tolist()[0])
            Table['completeness_metric'].append(comp_str(d['Completeness'].tolist()[0]))
            Table['contamination_metric'].append(con_str(d['Contamination'].tolist()[0]))
        else:
            Table['completeness'].append("NA")
            Table['contamination'].append("NA")
            Table['strain_heterogeneity'].append("NA")
            Table['size'].append("NA")
            Table['N50'].append("NA")
            Table['completeness_metric'].append("NA")
            Table['contamination_metric'].append("NA")

        # Add taxonomy info
        if wd.hasDb('Tdb'):
            Tdb = wd.get_db('Tdb')
            d = Tdb[Tdb['genome'] == row['genome']]
            Table['taxonomy'].append(d['taxonomy'][d['tax_confidence'] == d['tax_confidence'].max()].tolist()[0])
            Table['tax_confidence'].append(d['tax_confidence'].max())
        else:
            Table['taxonomy'].append('NA')
            Table['tax_confidence'].append('NA')

    Widb = pd.DataFrame(Table)
    return Widb

def run_tertiary_clustering(wd, **kwargs):
    # Create a new workdir inside data
    new_wd_loc = wd.get_dir('data') + 'tertiary_clustering'
    nWd = drep.WorkDirectory.WorkDirectory(new_wd_loc)

    # Make a copy of the kwargs
    kwargs_copy = kwargs.copy()
    if 'genomes' in kwargs_copy:
        del kwargs_copy['genomes']
    kwargs_copy['P_ani'] = kwargs_copy['P_ani'] - 0.05
    kwargs_copy['multiround_primary_clustering'] = False
    kwargs_copy['greedy_secondary_clustering'] = False

    # Figure out what genomes you're going to compare and make a new Bdb
    Bdb = wd.get_db('Bdb')
    Cdb = wd.get_db('Cdb')
    Wdb = wd.get_db('Wdb')
    BBdb = Bdb[Bdb['genome'].isin(Wdb['genome'].tolist())]
    nWd.store_db(BBdb, 'Bdb')

    # Copy over genome info for choose
    for name in ['genomeInfo', 'Chdb']:
        if wd.hasDb(name):
            nWd.store_db(wd.get_db(name), name)

    # Run the comparison
    drep.d_cluster.controller.d_cluster_wrapper(new_wd_loc, **kwargs_copy)

    # Reconsile results for Cdb
    nWd = drep.WorkDirectory.WorkDirectory(new_wd_loc)
    nCdb = nWd.get_db('Cdb')

    Cdb['original_secondary_cluster'] = Cdb['secondary_cluster']
    del Cdb['secondary_cluster']
    Cdb = pd.merge(Cdb, nCdb[['genome', 'secondary_cluster']], on='genome', how='left')

    rep_cdb = Cdb.dropna()
    old2new = rep_cdb.set_index('original_secondary_cluster')['secondary_cluster'].to_dict()
    Cdb['secondary_cluster'] = Cdb['original_secondary_cluster'].map(old2new)

    # Rename clusters
    old2new_names = {}
    for clust, db in Cdb.groupby('secondary_cluster'):
        if len(db['original_secondary_cluster'].unique()) == 1:
            old2new_names[clust] = db['original_secondary_cluster'].iloc[0]
        else:
            new_name =  f"{'.'.join(sorted(list(set([x.split('_')[0] for x in list(db['original_secondary_cluster'].unique())]))))}" + \
                        f"_{'.'.join(sorted(list(set([x.split('_')[1] for x in list(db['original_secondary_cluster'].unique())]))))}"
            old2new_names[clust] = new_name
            logging.debug(f"Clusters {list(db['original_secondary_cluster'].unique())} were merged into {new_name}")
    Cdb['secondary_cluster'] = Cdb['secondary_cluster'].map(old2new_names)

    # Store new Cdb
    wd.store_db(Cdb, 'Cdb')

    # Re-run choose
    drep.d_choose.d_choose_wrapper(wd.location, **kwargs_copy)
